import difflib
from collections.abc import Collection, Iterable, Iterator, Sequence
from itertools import islice
from typing import Any, TypeVar

from cognite_toolkit._cdf_tk.utils.file import yaml_safe_dump


def flatten_dict(dct: dict[str, Any]) -> dict[tuple[str, ...], Any]:
    """Flatten a dictionary to a list of tuples with the key path and value."""
    items: dict[tuple[str, ...], Any] = {}
    for key, value in dct.items():
        if isinstance(value, dict):
            for sub_key, sub_value in flatten_dict(value).items():
                items[(key, *sub_key)] = sub_value
        else:
            items[(key,)] = value
    return items


def to_diff(a: dict[str, Any], b: dict[str, Any]) -> Iterator[str]:
    a_str = yaml_safe_dump(a, sort_keys=True)
    b_str = yaml_safe_dump(b, sort_keys=True)

    return difflib.unified_diff(a_str.splitlines(), b_str.splitlines())


def in_dict(keys: Iterable[str], dictionary: dict) -> bool:
    return all(key in dictionary for key in keys)


def humanize_collection(collection: Collection[Any], /, *, sort: bool = True, bind_word: str = "and") -> str:
    if not collection:
        return ""
    elif len(collection) == 1:
        return str(next(iter(collection)))

    strings = (str(item) for item in collection)
    if sort:
        sequence = sorted(strings)
    else:
        sequence = list(strings)

    return f"{', '.join(sequence[:-1])} {bind_word} {sequence[-1]}"


def chunker(iterable: Iterable[Any], size: int) -> Iterator[list[Any]]:
    iterator = iter(iterable)
    while chunk := list(islice(iterator, size)):
        yield chunk


T_Sequence = TypeVar("T_Sequence", bound=Sequence)


def chunker_sequence(sequence: T_Sequence, size: int) -> Iterator[T_Sequence]:
    """Yield successive n-sized chunks from sequence."""
    for i in range(0, len(sequence), size):
        # MyPy does not expect sequence[i : i + size] to be of type T_Sequence
        yield sequence[i : i + size]  # type: ignore[misc]



from ._migrate import MigrateTimeseriesCommand, MigrationPrepareCommand
from ._populate import PopulateCommand
from ._profile import ProfileCommand
from ._purge import PurgeCommand
from .auth import AuthCommand
from .build_cmd import BuildCommand
from .clean import CleanCommand
from .collect import CollectCommand
from .deploy import DeployCommand
from .dump_data import DumpDataCommand
from .dump_resource import DumpResourceCommand
from .featureflag import FeatureFlagCommand
from .init import InitCommand
from .modules import ModulesCommand
from .pull import PullCommand
from .repo import RepoCommand
from .run import RunFunctionCommand, RunTransformationCommand, RunWorkflowCommand

__all__ = [
    "AuthCommand",
    "BuildCommand",
    "CleanCommand",
    "CollectCommand",
    "DeployCommand",
    "DumpDataCommand",
    "DumpResourceCommand",
    "FeatureFlagCommand",
    "InitCommand",
    "MigrateTimeseriesCommand",
    "MigrationPrepareCommand",
    "ModulesCommand",
    "PopulateCommand",
    "ProfileCommand",
    "PullCommand",
    "PurgeCommand",
    "RepoCommand",
    "RunFunctionCommand",
    "RunTransformationCommand",
    "RunWorkflowCommand",
]



import itertools
from collections import Counter, defaultdict

import pytest
from cognite.client.data_classes import (
    AssetList,
    AssetWrite,
    AssetWriteList,
    DataSetList,
    DataSetWrite,
    DataSetWriteList,
    LabelDefinitionList,
    LabelDefinitionWrite,
    RelationshipList,
    RelationshipWrite,
    RelationshipWriteList,
)
from cognite.client.data_classes.labels import LabelDefinitionWriteList

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.utils.cdf import label_aggregate_count, metadata_key_counts, relationship_aggregate_count


@pytest.fixture(scope="session")
def two_datasets(toolkit_client: ToolkitClient) -> DataSetList:
    datasets = DataSetWriteList(
        [
            DataSetWrite(
                name="Test Dataset 1",
                external_id="toolkit_test_metadata_key_counts_1",
                description="Test Dataset 1",
            ),
            DataSetWrite(
                name="Test Dataset 2",
                external_id="toolkit_test_metadata_key_counts_2",
                description="Test Dataset 2",
            ),
        ]
    )
    existing = toolkit_client.data_sets.retrieve_multiple(
        external_ids=datasets.as_external_ids(), ignore_unknown_ids=True
    )
    if missing := [dataset for dataset in datasets if dataset.external_id not in set(existing.as_external_ids())]:
        created = toolkit_client.data_sets.create(missing)
        existing.extend(created)
    return existing


@pytest.fixture(scope="session")
def two_hierarchies(toolkit_client: ToolkitClient, two_datasets: DataSetList) -> tuple[AssetList, AssetList]:
    hierarchies = [
        AssetWriteList(
            [
                AssetWrite(
                    name=f"Root Asset {i}",
                    external_id=f"toolkit_test_metadata_key_counts_root_asset_{i}",
                    metadata={f"metadata_key_{i}": "does-not-matter"},
                    data_set_id=two_datasets[i].id,
                ),
                AssetWrite(
                    name=f"Child Asset {i}",
                    parent_external_id=f"toolkit_test_metadata_key_counts_root_asset_{i}",
                    data_set_id=two_datasets[i].id,
                    external_id=f"toolkit_test_metadata_key_counts_child_asset_{i}",
                    metadata={
                        f"metadata_key_{i}": "does not matter either",
                        f"metadata_key_extra_{i}": "does not matter either",
                    },
                ),
                AssetWrite(
                    name=f"Child Asset extra {i}",
                    parent_external_id=f"toolkit_test_metadata_key_counts_root_asset_{i}",
                    external_id=f"toolkit_test_metadata_key_counts_child_asset_extra_{i}",
                    metadata={
                        f"extra_key{i}": "does not matter either",
                        f"another_extra_key{i}": "does not matter either",
                    },
                ),
            ]
        )
        for i in range(2)
    ]
    retrieved_hierarchies: list[AssetList] = []
    for hierarchy in hierarchies:
        existing = toolkit_client.assets.retrieve_multiple(
            external_ids=hierarchy.as_external_ids(), ignore_unknown_ids=True
        )
        if missing := [asset for asset in hierarchy if asset.external_id not in set(existing.as_external_ids())]:
            created = toolkit_client.assets.create(missing)
            existing.extend(created)
        retrieved_hierarchies.append(existing)
    return retrieved_hierarchies[0], retrieved_hierarchies[1]


@pytest.fixture(scope="session")
def asset_relationships(
    toolkit_client: ToolkitClient, two_hierarchies: tuple[AssetList, AssetList], two_datasets: DataSetList
) -> RelationshipList:
    dataset_id = two_datasets[0].id  # Using the first dataset for relationships
    hierarchy, _ = two_hierarchies  # Using the first hierarchy for relationships
    children_by_parent: dict[str, list[str]] = defaultdict(list)
    for asset in hierarchy:
        if asset.parent_external_id and asset.external_id:
            children_by_parent[asset.parent_external_id].append(asset.external_id)

    relationships = RelationshipWriteList(
        [
            RelationshipWrite(
                external_id=f"toolkit_test_relationship_count_{parent}_{child}",
                source_type="asset",
                source_external_id=parent,
                target_type="asset",
                target_external_id=child,
                data_set_id=dataset_id,
            )
            for parent, children in children_by_parent.items()
            for child in children
        ]
    )
    existing = toolkit_client.relationships.retrieve_multiple(
        external_ids=relationships.as_external_ids(), ignore_unknown_ids=True
    )
    if missing := [rel for rel in relationships if rel.external_id not in set(existing.as_external_ids())]:
        created = toolkit_client.relationships.create(missing)
        existing.extend(created)
    return existing


@pytest.fixture(scope="session")
def two_labels(toolkit_client: ToolkitClient, two_datasets: DataSetList) -> LabelDefinitionList:
    data_set_id = two_datasets[0].id  # Using the first dataset for labels
    labels = LabelDefinitionWriteList(
        [
            LabelDefinitionWrite(
                external_id="toolkit_test_label_aggregate_count_1",
                name="Test Label 1",
                description="This is a test label 1",
                data_set_id=data_set_id,
            ),
            LabelDefinitionWrite(
                external_id="toolkit_test_label_aggregate_count_2",
                name="Test Label 2",
                description="This is a test label 2",
                data_set_id=data_set_id,
            ),
        ]
    )
    existing = toolkit_client.labels.retrieve(external_id=labels.as_external_ids(), ignore_unknown_ids=True)
    if missing := [label for label in labels if label.external_id not in set(existing.as_external_ids())]:
        created = toolkit_client.labels.create(missing)
        existing.extend(created)
    return existing


class TestMetadataKeyCounts:
    def test_metadata_key_counts(self, toolkit_client: ToolkitClient) -> None:
        metadata_keys = metadata_key_counts(toolkit_client, "events")
        assert len(metadata_keys) > 0
        ill_formed = [
            (key, count)
            for key, count in metadata_keys
            if not isinstance(key, str) or not isinstance(count, int) or count < 0
        ]

        assert len(ill_formed) == 0, f"Ill-formed metadata keys: {ill_formed}"

    @pytest.mark.parametrize(
        "hierarchies, datasets",
        (
            (hierarchies, datasets)
            for hierarchies, datasets in itertools.product(
                [
                    None,
                    ["toolkit_test_metadata_key_counts_root_asset_0"],
                    ["toolkit_test_metadata_key_counts_root_asset_0", "toolkit_test_metadata_key_counts_root_asset_1"],
                ],
                [
                    None,
                    ["toolkit_test_metadata_key_counts_1"],
                    ["toolkit_test_metadata_key_counts_1", "toolkit_test_metadata_key_counts_2"],
                ],
            )
            if not (not hierarchies and not datasets)
        ),
        ids=(
            f"{hierarchy_str} and {dataset_str}"
            for hierarchy_str, dataset_str in itertools.product(
                ["no hierarchies", "one hierarchy", "two hierarchies"],
                ["no datasets", "one dataset", "two datasets"],
                # We filter out this case as it will return all assets in the project as it has no filtering.
            )
            if not (hierarchy_str == "no hierarchies" and dataset_str == "no datasets")
        ),
    )
    def test_metadata_key_counts_filtering_datasets_hierarchies(
        self,
        hierarchies: list[str] | None,
        datasets: list[str] | None,
        toolkit_client: ToolkitClient,
        two_hierarchies: tuple[AssetList, AssetList],
        two_datasets: DataSetList,
    ) -> None:
        hierarchy_external_to_internal = {assets[0].external_id: assets[0].id for assets in two_hierarchies}
        datasets_external_to_internal = {dataset.external_id: dataset.id for dataset in two_datasets}
        hierarchy_ids = (
            [hierarchy_external_to_internal[external_id] for external_id in hierarchies] if hierarchies else None
        )
        dataset_ids = [datasets_external_to_internal[external_id] for external_id in datasets] if datasets else None
        metadata_keys = metadata_key_counts(toolkit_client, "assets", hierarchies=hierarchy_ids, data_sets=dataset_ids)

        expected_keys = Counter()
        for hierarchy in two_hierarchies:
            for asset in hierarchy:
                if (hierarchy_ids is None or asset.root_id in hierarchy_ids) and (
                    dataset_ids is None or asset.data_set_id in dataset_ids
                ):
                    expected_keys.update(asset.metadata.keys())

        assert {key: count for key, count in metadata_keys} == dict(expected_keys.items())


class TestRelationshipAggregateCount:
    @pytest.mark.usefixtures("asset_relationships")
    def test_relationship_aggregate_count(self, toolkit_client: ToolkitClient) -> None:
        results = relationship_aggregate_count(toolkit_client)

        # We do not know how many relationship exists in the project, so we only check
        # that there is at least one relationship.
        total = sum(item.count for item in results)
        assert total > 0

    def test_relationship_aggregate_count_with_filtering(
        self, toolkit_client: ToolkitClient, asset_relationships: RelationshipList
    ) -> None:
        assert len(asset_relationships) > 0, "There should be some relationships to test with."
        data_set_id = asset_relationships[0].data_set_id
        assert data_set_id is not None, "The relationships should have a data set ID."
        results = relationship_aggregate_count(toolkit_client, [data_set_id])

        assert len(results) == 1
        item = results[0]
        assert item.source_type == "asset"
        assert item.target_type == "asset"
        assert item.count == len(asset_relationships)


class TestLabelAggregateCount:
    @pytest.mark.usefixtures("two_labels")
    def test_label_aggregate_count(self, toolkit_client: ToolkitClient) -> None:
        total = label_aggregate_count(toolkit_client)

        # We do not know how many labels exists in the project, so we only check
        # that there is at least one label.
        assert total > 0

    def test_label_aggregate_count_with_filtering(
        self, toolkit_client: ToolkitClient, two_labels: LabelDefinitionList
    ) -> None:
        assert len(two_labels) > 0, "There should be some labels to test with."
        data_set_id = two_labels[0].data_set_id
        assert data_set_id is not None, "The labels should have a data set ID."
        count = label_aggregate_count(toolkit_client, [data_set_id])

        assert count == len(two_labels)



# ruff: noqa: E402
import contextlib
import logging
import os
import platform
import shutil
import subprocess
import sys
from collections.abc import Iterator
from contextlib import contextmanager
from datetime import date
from pathlib import Path
from unittest.mock import patch

from cognite.client.config import global_config

from cognite_toolkit._cdf_tk.utils.file import safe_rmtree
from tests.auth_utils import EnvironmentVariables

# Do not warn the user about feature previews from the Cognite-SDK we use in Toolkit
global_config.disable_pypi_version_check = True
global_config.silence_feature_preview_warnings = True

from cognite_toolkit._cdf_tk.cdf_toml import CDFToml

# Hack to make the script work as running cdf modules upgrade
original_argv = sys.argv
sys.argv = ["cdf", "modules", "upgrade"]

import yaml
from dotenv import load_dotenv
from packaging.version import Version
from packaging.version import parse as parse_version
from rich import print
from rich.panel import Panel

from cognite_toolkit._cdf_tk.commands import BuildCommand, DeployCommand, ModulesCommand
from cognite_toolkit._cdf_tk.commands import _cli_commands as CLICommands
from cognite_toolkit._cdf_tk.commands._changes import ManualChange
from cognite_toolkit._cdf_tk.constants import ROOT_MODULES, SUPPORT_MODULE_UPGRADE_FROM_VERSION
from cognite_toolkit._cdf_tk.loaders import LOADER_BY_FOLDER_NAME
from cognite_toolkit._cdf_tk.utils import module_from_path
from cognite_toolkit._version import __version__

TEST_DIR_ROOT = Path(__file__).resolve().parent
PROJECT_INIT_DIR = TEST_DIR_ROOT / "project_inits"
PROJECT_INIT_DIR.mkdir(exist_ok=True)

TODAY = date.today()

logging.basicConfig(
    filename=f"module_upgrade_{TODAY.strftime('%Y-%m-%d')}.log",
    filemode="w",
    format="%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s",
    datefmt="%H:%M:%S",
    level=logging.INFO,
)


def run() -> None:
    from_earliest = len(original_argv) > 1 and "--earliest" in original_argv[1:3]
    from_latest = len(original_argv) > 1 and "--latest" in original_argv[1:3]
    selected_version = len(original_argv) > 1 and "--version" in original_argv[1:3] and original_argv[2]

    versions = get_versions_since(SUPPORT_MODULE_UPGRADE_FROM_VERSION)
    if selected_version:
        selected = parse_version(selected_version)
        versions = [version for version in versions if version == selected]
        if not versions:
            raise ValueError(f"Version {selected_version} is not available.")
    elif from_earliest and from_latest:
        versions = [versions[0], versions[-1]]
    elif from_earliest:
        versions = versions[-1:]
    elif from_latest:
        versions = versions[:1]

    for version in versions:
        create_project_init(str(version))

    print(
        Panel(
            "All projects inits created successfully.",
            expand=False,
            title="cdf-tk init executed for all past versions.",
        )
    )

    print(
        Panel(
            "Running module upgrade for all supported versions.",
            expand=False,
            title="cdf-tk module upgrade",
        )
    )
    for version in versions:
        with (
            local_tmp_project_path() as project_path,
            local_build_path() as build_path,
            get_env_vars() as env_vars,
        ):
            run_modules_upgrade(version, project_path, build_path, env_vars)


def get_versions_since(support_upgrade_from_version: str) -> list[Version]:
    result = subprocess.run("pip index versions cognite-toolkit --pre".split(), stdout=subprocess.PIPE)
    lines = result.stdout.decode().split("\n")
    for line in lines:
        if line.startswith("Available versions:"):
            raw_version_str = line.split(":", maxsplit=1)[1]
            supported_from = parse_version(support_upgrade_from_version)
            return [
                parsed
                for version in raw_version_str.split(",")
                if (parsed := parse_version(version.strip())) >= supported_from
            ]
    else:
        raise ValueError("Could not find available versions.")


def create_project_init(version: str) -> None:
    project_init = PROJECT_INIT_DIR / f"project_{version}"
    if project_init.exists():
        print(f"Project init for version {version} already exists.")
        return

    environment_directory = f".venv{version}"
    if (TEST_DIR_ROOT / environment_directory).exists():
        print(f"Environment for version {version} already exists")
    else:
        with chdir(TEST_DIR_ROOT):
            print(f"Creating environment for version {version}")
            create_venv = subprocess.run(["python", "-m", "venv", environment_directory])
            if create_venv.returncode != 0:
                raise ValueError(f"Failed to create environment for version {version}")

            if platform.system() == "Windows":
                install_toolkit = subprocess.run(
                    [f"{environment_directory}/Scripts/pip", "install", f"cognite-toolkit=={version}"]
                )
            else:
                install_toolkit = subprocess.run(
                    [f"{environment_directory}/bin/pip", "install", f"cognite-toolkit=={version}"]
                )

            if install_toolkit.returncode != 0:
                raise ValueError(f"Failed to install toolkit version {version}")

            if parse_version(version) < parse_version("0.2.4"):
                # Bug in pre 0.2.4 versions that was missing the packaging dependency
                if platform.system() == "Windows":
                    install_packing = subprocess.run(
                        [f"{environment_directory}/Scripts/pip", "install", "packaging==24.1"]
                    )
                else:
                    install_packing = subprocess.run([f"{environment_directory}/bin/pip", "install", "packaging==24.1"])

                if install_packing.returncode != 0:
                    raise ValueError(f"Failed to install packing version {version}")

            print(f"Environment for version {version} created")

    modified_env_variables = os.environ.copy()
    repo_root = TEST_DIR_ROOT.parent
    if "PYTHONPATH" in modified_env_variables:
        # Need to remove the repo root from PYTHONPATH to avoid importing the wrong version of the toolkit
        # (This is typically set by the IDE, for example, PyCharm sets it when running tests).
        modified_env_variables["PYTHONPATH"] = modified_env_variables["PYTHONPATH"].replace(str(repo_root), "")
    if platform.system() == "Windows":
        old_version_script_dir = Path(f"{environment_directory}/Scripts/")
    else:
        old_version_script_dir = Path(f"{environment_directory}/bin/")
    with chdir(TEST_DIR_ROOT):
        version_parsed = parse_version(version)
        if version_parsed >= parse_version("0.3.0a1"):
            cmd = [
                str(old_version_script_dir / "cdf"),
                "modules",
                "init",
                f"{PROJECT_INIT_DIR.name}/{project_init.name}",
                "--clean",
                "--all",
            ]
        else:
            cmd = [
                str(old_version_script_dir / "cdf-tk"),
                "init",
                f"{PROJECT_INIT_DIR.name}/{project_init.name}",
            ]

        output = subprocess.run(
            cmd,
            capture_output=True,
            shell=True if platform.system() == "Windows" else False,
            env=modified_env_variables,
        )
        if output.returncode != 0:
            print(output.stderr.decode())
            raise ValueError(f"Failed to create project init for version {version}.")

        cdf_toml_path = TEST_DIR_ROOT / CDFToml.file_name
        if cdf_toml_path.exists():
            shutil.move(cdf_toml_path, project_init / CDFToml.file_name)

    print(f"Project init for version {version} created.")
    with chdir(TEST_DIR_ROOT):
        safe_rmtree(environment_directory)


def run_modules_upgrade(
    previous_version: Version, project_path: Path, build_path: Path, env_vars: EnvironmentVariables
) -> None:
    if (TEST_DIR_ROOT / CDFToml.file_name).exists():
        # Cleanup after previous run
        (TEST_DIR_ROOT / CDFToml.file_name).unlink()

    project_init = PROJECT_INIT_DIR / f"project_{previous_version!s}"
    # Copy the project to a temporary location as the upgrade command modifies the project.
    shutil.copytree(project_init, project_path, dirs_exist_ok=True)
    if previous_version >= parse_version("0.3.0a1"):
        # Move out the CDF.toml file to use
        shutil.move(project_path / CDFToml.file_name, TEST_DIR_ROOT / CDFToml.file_name)

    with chdir(TEST_DIR_ROOT):
        modules = ModulesCommand(print_warning=False)
        # This is to allow running the function with having uncommitted changes in the repository.
        with patch.object(CLICommands, "has_uncommitted_changes", lambda: False):
            changes = modules.upgrade(project_path)
        logging.info(f"Changes for version {previous_version!s} to {__version__}: {len(changes)}")
        delete_modules_requiring_manual_changes(changes)

        # Update the config file to run include all modules.
        update_config_yaml_to_select_all_modules(project_path)

        if previous_version < parse_version("0.2.0a4"):
            # Bug in pre 0.2.0a4 versions
            pump_view = (
                project_path
                / "cognite_modules"
                / "experimental"
                / "example_pump_data_model"
                / "data_models"
                / "4.Pump.view.yaml"
            )
            if pump_view.exists():
                pump_view.write_text(pump_view.read_text().replace("external_id", "externalId"))

        build = BuildCommand(print_warning=False)
        build.execute(False, project_path, build_path, selected=None, build_env_name="dev", no_clean=False)

        deploy = DeployCommand(print_warning=False)
        deploy.deploy_build_directory(
            env_vars,
            build_path,
            build_env_name="dev",
            dry_run=True,
            drop=False,
            drop_data=False,
            force_update=False,
            include=list(LOADER_BY_FOLDER_NAME),
            verbose=False,
        )

    print(
        Panel(
            f"Module upgrade for version {previous_version!s} to {__version__} completed successfully.",
            expand=False,
            style="green",
        )
    )
    logging.info(f"Module upgrade for version {previous_version!s} to {__version__} completed successfully.")


def delete_modules_requiring_manual_changes(changes):
    for change in changes:
        if not isinstance(change, ManualChange):
            continue
        for file in change.needs_to_change():
            if file.is_dir():
                safe_rmtree(file)
            else:
                module = module_from_path(file)
                for part in reversed(file.parts):
                    if part == module:
                        break
                    file = file.parent
                if file.exists():
                    safe_rmtree(file)


def update_config_yaml_to_select_all_modules(project_path):
    config_yaml = project_path / "config.dev.yaml"
    assert config_yaml.exists()
    yaml_data = yaml.safe_load(config_yaml.read_text())
    yaml_data["environment"]["selected"] = []
    for root_module in ROOT_MODULES:
        if (project_path / root_module).exists() and any(
            yaml_file for yaml_file in (project_path / root_module).rglob("*.yaml")
        ):
            yaml_data["environment"]["selected"].append(f"{root_module}/")
    config_yaml.write_text(yaml.dump(yaml_data))


@contextlib.contextmanager
def chdir(new_dir: Path) -> Iterator[None]:
    """
    Change directory to new_dir and return to the original directory when exiting the context.

    Args:
        new_dir: The new directory to change to.

    """
    current_working_dir = Path.cwd()
    os.chdir(new_dir)

    try:
        yield

    finally:
        os.chdir(current_working_dir)


@contextmanager
def get_env_vars() -> Iterator[EnvironmentVariables]:
    load_dotenv(TEST_DIR_ROOT.parent / ".env")

    try:
        yield EnvironmentVariables.create_from_environ()
    finally:
        ...


@contextmanager
def local_tmp_project_path() -> Path:
    project_path = TEST_DIR_ROOT / "tmp-project"
    if project_path.exists():
        safe_rmtree(project_path)
    project_path.mkdir(exist_ok=True)
    try:
        yield project_path
    finally:
        ...


@contextmanager
def local_build_path() -> Path:
    build_path = TEST_DIR_ROOT / "build"
    if build_path.exists():
        safe_rmtree(build_path)

    build_path.mkdir(exist_ok=True)
    # This is a small hack to get 0.1.0b1-4 working
    (build_path / "file.txt").touch(exist_ok=True)
    try:
        yield build_path
    finally:
        ...


if __name__ == "__main__":
    run()



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes import SpaceYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def invalid_space_test_cases() -> Iterable:
    yield pytest.param(
        {"externalId": "mySpace"},
        {"Unused field: 'externalId'", "Missing required field: 'space'"},
        id="Missing required field: space",
    )
    yield pytest.param(
        {"space": "space"},
        {
            "In field space 'space' is a reserved space. Reserved Spaces: cdf, dms, edge, "
            "node, pg3, shared, space and system"
        },
        id="Reserved name",
    )
    yield pytest.param(
        {
            "space": "invalid<characters>",
        },
        {"In field space string should match pattern '^[a-zA-Z][a-zA-Z0-9_-]{0,41}[a-zA-Z0-9]?$'"},
        id="Invalid characters in space name",
    )


class TestSpaceYAML:
    @pytest.mark.parametrize("data", list(find_resources("Space")))
    def test_load_valid_space(self, data: dict[str, object]) -> None:
        loaded = SpaceYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_space_test_cases()))
    def test_invalid_asset_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, SpaceYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



from abc import ABC, abstractmethod
from typing import ClassVar, Literal

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.utils.cdf import (
    label_aggregate_count,
    label_count,
    metadata_key_counts,
    relationship_aggregate_count,
)


class AssetCentricAggregator(ABC):
    _transformation_destination: ClassVar[tuple[str, ...]]

    def __init__(self, client: ToolkitClient) -> None:
        self.client = client

    @property
    @abstractmethod
    def display_name(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def count(self) -> int:
        raise NotImplementedError

    def transformation_count(self) -> int:
        """Returns the number of transformations associated with the resource."""
        transformation_count = 0
        for destination in self._transformation_destination:
            for chunk in self.client.transformations(chunk_size=1000, destination_type=destination, limit=None):
                transformation_count += len(chunk)
        return transformation_count


class MetadataAggregator(AssetCentricAggregator, ABC):
    def __init__(
        self, client: ToolkitClient, resource_name: Literal["assets", "events", "files", "timeseries", "sequences"]
    ) -> None:
        super().__init__(client)
        self.resource_name = resource_name

    def metadata_key_count(self) -> int:
        return len(metadata_key_counts(self.client, self.resource_name))


class LabelAggregator(MetadataAggregator, ABC):
    def label_count(self) -> int:
        return len(label_count(self.client, self.resource_name))


class AssetAggregator(LabelAggregator):
    _transformation_destination = ("assets", "asset_hierarchy")

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client, "assets")

    @property
    def display_name(self) -> str:
        return "Assets"

    def count(self) -> int:
        return self.client.assets.aggregate_count()


class EventAggregator(MetadataAggregator):
    _transformation_destination = ("events",)

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client, "events")

    @property
    def display_name(self) -> str:
        return "Events"

    def count(self) -> int:
        return self.client.events.aggregate_count()


class FileAggregator(LabelAggregator):
    _transformation_destination = ("files",)

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client, "files")

    @property
    def display_name(self) -> str:
        return "Files"

    def count(self) -> int:
        response = self.client.files.aggregate()
        if response:
            return response[0].count
        else:
            return 0


class TimeSeriesAggregator(MetadataAggregator):
    _transformation_destination = ("timeseries",)

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client, "timeseries")

    @property
    def display_name(self) -> str:
        return "TimeSeries"

    def count(self) -> int:
        return self.client.time_series.aggregate_count()


class SequenceAggregator(MetadataAggregator):
    _transformation_destination = ("sequences",)

    def __init__(self, client: ToolkitClient) -> None:
        super().__init__(client, "sequences")

    @property
    def display_name(self) -> str:
        return "Sequences"

    def count(self) -> int:
        return self.client.sequences.aggregate_count()


class RelationshipAggregator(AssetCentricAggregator):
    _transformation_destination = ("relationships",)

    @property
    def display_name(self) -> str:
        return "Relationships"

    def count(self) -> int:
        results = relationship_aggregate_count(self.client)
        return sum(result.count for result in results)


class LabelCountAggregator(AssetCentricAggregator):
    _transformation_destination = ("labels",)

    @property
    def display_name(self) -> str:
        return "Labels"

    def count(self) -> int:
        return label_aggregate_count(self.client)



from pathlib import Path

import questionary
from cognite.client.data_classes.capabilities import (
    Capability,
    DataModelInstancesAcl,
    DataModelsAcl,
    DataSetScope,
    SpaceIDScope,
    TimeSeriesAcl,
)
from cognite.client.data_classes.data_modeling import DirectRelationReference, ViewId
from cognite.client.data_classes.data_modeling.cdm.v1 import CogniteTimeSeriesApply
from cognite.client.exceptions import CogniteAPIError
from rich import print
from rich.panel import Panel
from rich.progress import track

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.extended_timeseries import ExtendedTimeSeries
from cognite_toolkit._cdf_tk.commands._base import ToolkitCommand
from cognite_toolkit._cdf_tk.constants import DMS_INSTANCE_LIMIT_MARGIN
from cognite_toolkit._cdf_tk.exceptions import (
    AuthenticationError,
    ToolkitValueError,
)
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils.collection import chunker_sequence

from .data_classes import MigrationMappingList


class MigrateTimeseriesCommand(ToolkitCommand):
    cdf_cdm = "cdf_cdm"
    view_id = ViewId(cdf_cdm, "CogniteTimeSeries", "v1")
    cdf_cdm_units = "cdf_cdm_units"
    schema_spaces = frozenset({cdf_cdm, cdf_cdm_units})
    chunk_size = 1000

    def migrate_timeseries(
        self,
        client: ToolkitClient,
        mapping_file: Path,
        dry_run: bool = False,
        verbose: bool = False,
        auto_yes: bool = False,
    ) -> None:
        """Migrate resources from Asset-Centric to data modeling in CDF."""
        mappings = MigrationMappingList.read_mapping_file(mapping_file)

        self._validate_access(client, mappings)
        self._validate_timeseries_existence(client, mappings)
        self._validate_available_capacity(client, mappings)

        if dry_run:
            self.console(f"Dry run mode. Would have migrated {len(mappings):,} TimeSeries to CogniteTimeSeries.")
            return
        if not auto_yes and self._confirm(mappings) is False:
            return
        self._migrate(client, mappings, verbose)

    def _validate_access(self, client: ToolkitClient, mappings: MigrationMappingList) -> None:
        required_capabilities: list[Capability] = [
            DataModelsAcl(actions=[DataModelsAcl.Action.Read], scope=SpaceIDScope(list(self.schema_spaces))),
            DataModelInstancesAcl(
                actions=[
                    DataModelInstancesAcl.Action.Read,
                    DataModelInstancesAcl.Action.Write,
                    DataModelInstancesAcl.Action.Write_Properties,
                ],
                scope=SpaceIDScope(list(mappings.spaces())),
            ),
        ]
        if data_set_ids := mappings.get_data_set_ids():
            required_capabilities.append(
                TimeSeriesAcl(
                    actions=[TimeSeriesAcl.Action.Read, TimeSeriesAcl.Action.Write],
                    scope=DataSetScope(list(data_set_ids)),
                )
            )
        if missing := client.iam.verify_capabilities(required_capabilities):
            raise AuthenticationError(f"Missing required capabilities: {humanize_collection(missing)}.")

    def _validate_timeseries_existence(self, client: ToolkitClient, mappings: MigrationMappingList) -> None:
        total_validated = 0
        chunk: MigrationMappingList
        for chunk in track(  # type: ignore[assignment]
            chunker_sequence(mappings, size=self.chunk_size),
            description="Validating...",
            total=len(mappings) // self.chunk_size + 1,
        ):
            try:
                timeseries = client.time_series.retrieve_multiple(
                    ids=chunk.get_ids(),
                    external_ids=chunk.get_external_ids(),
                    ignore_unknown_ids=True,
                )
            except CogniteAPIError as e:
                raise AuthenticationError(
                    f"Failed to retrieve TimeSeries. This is likely due to lack of permissions: {e!s}"
                ) from e

            missing_count = len(timeseries) - len(mappings)
            if missing_count > 0:
                raise ToolkitValueError(f"Missing {missing_count} TimeSeries does not exist in CDF.")

            existing_result = client.data_modeling.instances.retrieve(chunk.as_node_ids())
            if len(existing_result.nodes) != 0:
                raise ToolkitValueError(
                    "Some of the TimeSeries you are trying to migrate already exist in Data Modeling. "
                    f"Please remove the following TimeSeries from the mapping file {humanize_collection(existing_result.nodes.as_ids())}"
                )
            total_validated += len(timeseries)
        print(
            f"Validated {total_validated:,} TimeSeries for migration. "
            f"{len(mappings):,} mappings provided in the mapping file."
        )

    def _validate_available_capacity(self, client: ToolkitClient, mappings: MigrationMappingList) -> None:
        """Validate that the project has enough capacity to accommodate the migration."""
        try:
            stats = client.data_modeling.statistics.project()
        except CogniteAPIError:
            # This endpoint is not yet in alpha, it may change or not be available.
            self.warn(HighSeverityWarning("Cannot check the instances capacity proceeding with migration anyway."))
            return
        available_capacity = stats.instances.instances_limit - stats.instances.instances
        available_capacity_after = available_capacity - len(mappings)

        if available_capacity_after < DMS_INSTANCE_LIMIT_MARGIN:
            raise ToolkitValueError(
                "Cannot proceed with migration, not enough instance capacity available. Total capacity after migration"
                f"would be {available_capacity_after:,} instances, which is less than the required margin of"
                f"{DMS_INSTANCE_LIMIT_MARGIN:,} instances. Please increase the instance capacity in your CDF project"
                f" or delete some existing instances before proceeding with the migration of {len(mappings):,} timeseries."
            )
        total_instances = stats.instances.instances + len(mappings)
        self.console(
            f"Project has enough capacity for migration. Total instances after migration: {total_instances:,}."
        )

    @staticmethod
    def _confirm(mappings: MigrationMappingList) -> bool:
        print(
            Panel(
                f"[red]WARNING:[/red] This operation [bold]cannot be undone[/bold]! "
                f"{len(mappings):,} TimeSeries will linked to the new CogniteTimeSeries. "
                "This linking cannot be undone",
                style="bold",
                title="Migrate asset-centric TimeSeries to CogniteTimeSeries",
                title_align="left",
                border_style="red",
                expand=False,
            )
        )

        if not questionary.confirm("Are you really sure you want to continue?", default=False).ask():
            print("Migration cancelled by user.")
            return False
        return True

    def _migrate(self, client: ToolkitClient, mappings: MigrationMappingList, verbose: bool) -> None:
        print("Migrating TimeSeries to CogniteTimeSeries...")
        total_migrated = 0
        for chunk in track(
            chunker_sequence(mappings, size=self.chunk_size),
            description="Migrating TimeSeries to CogniteTimeSeries...",
            total=len(mappings) // self.chunk_size + 1,
        ):  # type: ignore[assignment]
            if verbose:
                print(f"Migrating {len(chunk):,} TimeSeries...")

            # Set pending IDs for the chunk of mappings
            try:
                pending_timeseries = client.time_series.set_pending_ids(chunk.as_pending_ids())
            except CogniteAPIError as e:
                raise ToolkitValueError(f"Failed to set pending IDs for TimeSeries: {e!s}") from e

            # The ExtendedTimeSeriesList is iterating ExtendedTimeSeries objects.
            converted_timeseries = [self.as_cognite_timeseries(ts) for ts in pending_timeseries]  # type: ignore[arg-type]
            try:
                created = client.data_modeling.instances.apply_fast(converted_timeseries)
            except CogniteAPIError as e:
                raise ToolkitValueError(f"Failed to apply TimeSeries: {e!s}") from e
            if verbose:
                print(f"Created {len(created):,} CogniteTimeSeries.")
            total_migrated += len(created)
        print(f"Successfully migrated {total_migrated:,} TimeSeries to CogniteTimeSeries.")

    @classmethod
    def as_cognite_timeseries(cls, ts: ExtendedTimeSeries) -> CogniteTimeSeriesApply:
        if ts.pending_instance_id is None:
            raise ToolkitValueError("ExtendedTimeSeries must have a pending_instance_id set before migration.")
        if ts.is_step is None:
            raise ToolkitValueError("ExtendedTimeSeries must have is_step set before migration.")

        return CogniteTimeSeriesApply(
            space=ts.pending_instance_id.space,
            external_id=ts.pending_instance_id.external_id,
            name=ts.name,
            description=ts.description,
            is_step=ts.is_step,
            time_series_type="string" if ts.is_string else "numeric",
            source_unit=ts.unit,
            unit=DirectRelationReference(cls.cdf_cdm_units, ts.unit_external_id) if ts.unit_external_id else None,
        )



from typing import Dict, Union

import orjson
from requests import Response, Session, adapters


class IceCreamFactoryAPI:
    """Class for Ice Cream Factory API."""

    def __init__(self, base_url: str):
        self.base_url = base_url
        self.adapter = adapters.HTTPAdapter(max_retries=3)
        self.session = Session()
        self.session.mount("https://", self.adapter)

    def get_response(
        self, headers: Dict[str, str], url_suffix: str, params: Dict[str, Union[str, int, float]] = {}
    ) -> Response:
        """
        Get response from API.

        Args:
            headers: request header
            url_suffix: string to add to base url
            params: query parameters
        """

        response = self.session.get(f"{self.base_url}/{url_suffix}", headers=headers, timeout=40, params=params)
        response.raise_for_status()
        return response

    def get_datapoints(self, timeseries_ext_id: str, start: Union[str, int, float], end: Union[str, int, float]):
        """
        Get datapoints for a timeseries external id. This will also return datapoints for an associated timeseries

        (e.g. request for external id "HPM2C561:planned_status" will return datapoints for "HPM2C561:planned_status" AND
        "HPM2C561:status". Similar, request for timeseries with external id "HPM2C561:count" will return datapoints for
        "HPM2C561:count" AND ""HPM2C561:good").

        Args:
            timeseries_ext_id: external id of timeseries to get datapoints for
            start: start for datapoints (UNIX timestamp (int, float) or string with format 'YYYY-MM-DD HH:MM')
            end: end for datapoints (UNIX timestamp (int, float) or string with format 'YYYY-MM-DD HH:MM')
        """
        params = {"start": start, "end": end, "external_id": timeseries_ext_id}
        response = self.get_response(headers={}, url_suffix="datapoints/oee", params=params)

        response_dict = orjson.loads(response.content)

        response_dict = [
            {
                "instance_id": ts,
                "datapoints": [
                    # convert timestamp to ms (*1000) for CDF uploads
                    {"timestamp": dp[0] * 1000, "value": dp[1]}
                    for dp in dps
                ]
            } for ts, dps in response_dict.items() if len(dps) > 1
        ]

        return response_dict


# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import itertools
from collections import defaultdict
from collections.abc import Hashable, Iterable, Sequence
from pathlib import Path
from typing import Any, cast, final

from cognite.client.data_classes.capabilities import (
    Capability,
    RawAcl,
)
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils.useful_types import SequenceNotStr
from rich import print
from rich.console import Console

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.raw import RawDatabase, RawDatabaseList, RawTable, RawTableList
from cognite_toolkit._cdf_tk.loaders._base_loaders import ResourceContainerLoader, ResourceLoader
from cognite_toolkit._cdf_tk.resource_classes import DatabaseYAML, TableYAML

from .auth_loaders import GroupAllScopedLoader


@final
class RawDatabaseLoader(
    ResourceContainerLoader[RawDatabase, RawDatabase, RawDatabase, RawDatabaseList, RawDatabaseList]
):
    item_name = "raw tables"
    folder_name = "raw"
    filename_pattern = r"^(?!.*Table$).*$"
    resource_cls = RawDatabase
    resource_write_cls = RawDatabase
    list_cls = RawDatabaseList
    list_write_cls = RawDatabaseList
    kind = "Database"
    yaml_cls = DatabaseYAML
    dependencies = frozenset({GroupAllScopedLoader})
    support_update = False
    _doc_url = "Raw/operation/createDBs"

    def __init__(self, client: ToolkitClient, build_dir: Path, console: Console | None):
        super().__init__(client, build_dir, console)
        self._loaded_db_names: set[str] = set()

    @property
    def display_name(self) -> str:
        return "raw databases"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[RawDatabase] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [RawAcl.Action.Read, RawAcl.Action.List]
            if read_only
            else [RawAcl.Action.Read, RawAcl.Action.Write, RawAcl.Action.List]
        )

        scope: RawAcl.Scope.All | RawAcl.Scope.Table = RawAcl.Scope.All()  # type: ignore[valid-type]
        if items:
            tables_by_database: dict[str, list[str]] = {}
            for item in items:
                tables_by_database[item.db_name] = []

            scope = RawAcl.Scope.Table(dict(tables_by_database)) if tables_by_database else RawAcl.Scope.All()  # type: ignore[arg-type]

        return RawAcl(actions, scope)  # type: ignore[arg-type]

    @classmethod
    def get_id(cls, item: RawDatabase | dict) -> RawDatabase:
        if isinstance(item, dict):
            return RawDatabase(item["dbName"])
        return item

    @classmethod
    def dump_id(cls, id: RawDatabase) -> dict[str, Any]:
        return {"dbName": id.db_name}

    def create(self, items: RawDatabaseList) -> RawDatabaseList:
        database_list = self.client.raw.databases.create([db.db_name for db in items])
        return RawDatabaseList([RawDatabase(db_name=db.name) for db in database_list if db.name])

    def retrieve(self, ids: SequenceNotStr[RawDatabase]) -> RawDatabaseList:
        database_list = self.client.raw.databases.list(limit=-1)
        target_dbs = {db.db_name for db in ids}
        return RawDatabaseList([RawDatabase(db_name=db.name) for db in database_list if db.name in target_dbs])

    def delete(self, ids: SequenceNotStr[RawDatabase]) -> int:
        db_names = [table.db_name for table in ids]
        try:
            self.client.raw.databases.delete(db_names)
        except CogniteAPIError as e:
            # Bug in API, missing is returned as failed
            if e.failed and (db_names := [name for name in db_names if name not in e.failed]):
                self.client.raw.databases.delete(db_names)
            elif e.code == 404 and "not found" in e.message and "database" in e.message:
                return 0
            else:
                raise e
        return len(db_names)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[RawDatabase]:
        return (RawDatabase(db_name=cast(str, db.name)) for db in self.client.raw.databases)

    def count(self, ids: SequenceNotStr[RawDatabase]) -> int:
        nr_of_tables = 0
        for db_name, raw_tables in itertools.groupby(sorted(ids, key=lambda x: x.db_name), key=lambda x: x.db_name):
            try:
                tables = self.client.raw.tables.list(db_name=db_name, limit=-1)
            except CogniteAPIError as e:
                if db_name in {item.get("name") for item in e.missing or []}:
                    continue
                raise e
            nr_of_tables += len(tables.data)
        return nr_of_tables

    def drop_data(self, ids: SequenceNotStr[RawDatabase]) -> int:
        nr_of_tables = 0
        for db_name, raw_tables in itertools.groupby(sorted(ids, key=lambda x: x.db_name), key=lambda x: x.db_name):
            try:
                existing = self.client.raw.tables.list(db_name=db_name, limit=-1).as_names()
            except CogniteAPIError as e:
                if db_name in {item.get("name") for item in e.missing or []}:
                    continue
                raise e
            if existing:
                self.client.raw.tables.delete(db_name=db_name, name=existing)
                nr_of_tables += len(existing)
        return nr_of_tables


@final
class RawTableLoader(ResourceContainerLoader[RawTable, RawTable, RawTable, RawTableList, RawTableList]):
    item_name = "raw rows"
    folder_name = "raw"
    filename_pattern = r"^(?!.*Database$).*$"
    resource_cls = RawTable
    resource_write_cls = RawTable
    list_cls = RawTableList
    list_write_cls = RawTableList
    kind = "Table"
    yaml_cls = TableYAML
    support_update = False
    dependencies = frozenset({RawDatabaseLoader, GroupAllScopedLoader})
    _doc_url = "Raw/operation/createTables"
    parent_resource = frozenset({RawDatabaseLoader})

    def __init__(self, client: ToolkitClient, build_dir: Path, console: Console | None):
        super().__init__(client, build_dir, console)
        self._printed_warning = False

    @property
    def display_name(self) -> str:
        return "raw tables"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[RawTable] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [RawAcl.Action.Read, RawAcl.Action.List]
            if read_only
            else [RawAcl.Action.Read, RawAcl.Action.Write, RawAcl.Action.List]
        )

        scope: RawAcl.Scope.All | RawAcl.Scope.Table = RawAcl.Scope.All()  # type: ignore[valid-type]
        if items:
            tables_by_database = defaultdict(list)
            for item in items:
                tables_by_database[item.db_name].append(item.table_name)

            scope = RawAcl.Scope.Table(dict(tables_by_database)) if tables_by_database else RawAcl.Scope.All()  # type: ignore[arg-type]

        return RawAcl(actions, scope)  # type: ignore[arg-type]

    @classmethod
    def get_id(cls, item: RawTable | dict) -> RawTable:
        if isinstance(item, dict):
            if missing := tuple(k for k in {"dbName", "tableName"} if k not in item):
                # We need to raise a KeyError with all missing keys to get the correct error message.
                raise KeyError(*missing)
            return RawTable(item["dbName"], item["tableName"])
        return item

    @classmethod
    def dump_id(cls, id: RawTable) -> dict[str, Any]:
        return {"dbName": id.db_name, "tableName": id.table_name}

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceLoader], Hashable]]:
        if "dbName" in item:
            yield RawDatabaseLoader, RawDatabase(item["dbName"])

    def create(self, items: RawTableList) -> RawTableList:
        created = RawTableList([])
        for db_name, raw_tables in itertools.groupby(sorted(items, key=lambda x: x.db_name), key=lambda x: x.db_name):
            tables = [table.table_name for table in raw_tables]
            new_tables = self.client.raw.tables.create(db_name=db_name, name=tables)
            created.extend([RawTable(db_name=db_name, table_name=cast(str, table.name)) for table in new_tables])
        return created

    def retrieve(self, ids: SequenceNotStr[RawTable]) -> RawTableList:
        retrieved = RawTableList([])
        for db_name, raw_tables in itertools.groupby(sorted(ids, key=lambda x: x.db_name), key=lambda x: x.db_name):
            expected_tables = {table.table_name for table in raw_tables}
            try:
                tables = self.client.raw.tables.list(db_name=db_name, limit=-1)
            except CogniteAPIError as e:
                if db_name in {item.get("name") for item in e.missing or []}:
                    continue
                raise e
            retrieved.extend(
                [RawTable(db_name=db_name, table_name=table.name) for table in tables if table.name in expected_tables]
            )
        return retrieved

    def delete(self, ids: SequenceNotStr[RawTable]) -> int:
        count = 0
        for db_name, raw_tables in itertools.groupby(sorted(ids, key=lambda x: x.db_name), key=lambda x: x.db_name):
            tables = [table.table_name for table in raw_tables if table.table_name]
            if tables:
                try:
                    self.client.raw.tables.delete(db_name=db_name, name=tables)
                except CogniteAPIError as e:
                    if e.code != 404:
                        raise e
                    # Missing is returned as failed
                    missing = {item.get("name") for item in (e.missing or [])}.union(set(e.failed or []))
                    if "not found" in e.message and "database" in e.message:
                        continue
                    elif tables := [name for name in tables if name not in missing]:
                        self.client.raw.tables.delete(db_name=db_name, name=tables)
                    elif not tables:
                        # Table does not exist.
                        continue
                    else:
                        raise e
                count += len(tables)
        return count

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[RawTable]:
        for parent_id in parent_ids or (RawDatabase(cast(str, db.name)) for db in self.client.raw.databases):
            if not isinstance(parent_id, RawDatabase):
                continue
            for table in self.client.raw.tables(cast(str, parent_id.db_name)):
                yield RawTable(db_name=cast(str, parent_id.db_name), table_name=cast(str, table.name))

    def count(self, ids: SequenceNotStr[RawTable]) -> int:
        if not self._printed_warning:
            print("  [bold green]INFO:[/] Raw rows do not support count (there is no aggregation method).")
            self._printed_warning = True
        return -1

    def drop_data(self, ids: SequenceNotStr[RawTable]) -> int:
        for db_name, raw_tables in itertools.groupby(sorted(ids, key=lambda x: x.db_name), key=lambda x: x.db_name):
            try:
                existing = set(self.client.raw.tables.list(db_name=db_name, limit=-1).as_names())
            except CogniteAPIError as e:
                if db_name in {item.get("name") for item in e.missing or []}:
                    continue
                raise e
            tables = [table.table_name for table in raw_tables if table.table_name in existing]
            if tables:
                self.client.raw.tables.delete(db_name=db_name, name=tables)
        return -1



from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.containers import BTreeIndex

SPACE = dm.SpaceApply(
    "cognite_migration", description="Space for the asset-centric to data modeling migration", name="cdf_migration"
)
MAPPING_CONTAINER = dm.ContainerApply(
    space=SPACE.space,
    external_id="Mapping",
    used_for="node",
    properties={
        "resourceType": dm.ContainerProperty(
            type=dm.data_types.Enum(
                values={
                    "timeseries": dm.data_types.EnumValue(),
                    "asset": dm.data_types.EnumValue(),
                    "file": dm.data_types.EnumValue(),
                    "event": dm.data_types.EnumValue(),
                    "sequence": dm.data_types.EnumValue(),
                }
            ),
            nullable=False,
        ),
        "id": dm.ContainerProperty(
            type=dm.data_types.Int64(),
            nullable=False,
        ),
        "dataSetId": dm.ContainerProperty(
            type=dm.data_types.Int64(),
            nullable=True,
        ),
        "classicExternalId": dm.ContainerProperty(
            type=dm.data_types.Text(),
            nullable=True,
        ),
    },
    indexes={
        "id": BTreeIndex(["id"], cursorable=True),
        "resourceType": BTreeIndex(["resourceType", "id"], cursorable=False),
    },
)

MAPPING_VIEW = dm.ViewApply(
    space=SPACE.space,
    external_id="Mapping",
    version="v1",
    name="Mapping",
    description="The mapping between asset-centric and data modeling resources",
    properties={
        "resourceType": dm.MappedPropertyApply(
            container=MAPPING_CONTAINER.as_id(),
            container_property_identifier="resourceType",
        ),
        "id": dm.MappedPropertyApply(
            container=MAPPING_CONTAINER.as_id(),
            container_property_identifier="id",
        ),
        "dataSetId": dm.MappedPropertyApply(
            container=MAPPING_CONTAINER.as_id(),
            container_property_identifier="dataSetId",
        ),
        "classicExternalId": dm.MappedPropertyApply(
            container=MAPPING_CONTAINER.as_id(),
            container_property_identifier="classicExternalId",
        ),
    },
)

COGNITE_MIGRATION_MODEL = dm.DataModelApply(
    space=SPACE.space,
    external_id="CogniteMigration",
    version="v1",
    name="CDF Migration Model",
    description="Data model for migrating asset-centric resources to data modeling resources in CDF.",
    views=[MAPPING_VIEW.as_id()],
)

MODEL_ID = COGNITE_MIGRATION_MODEL.as_id()



"""Creating new class for the SequenceRows and SequenceRowsList to be used in the toolkit. This is because the
SequenceRows and SequenceRowsList classes in the cognite-sdk-python are not compatible with the CDF API.

The new classes includes Writable version with the .as_write() method as well that the write version of the rows
only need a list of strings for the columns and not the entire column object.
"""

import collections.abc
from typing import Any

from cognite.client import CogniteClient
from cognite.client.data_classes._base import (
    CogniteResource,
    CogniteResourceList,
    IdTransformerMixin,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)
from cognite.client.data_classes.sequences import SequenceColumn, SequenceColumnList, SequenceRow, SequenceRows
from cognite.client.utils._auxiliary import at_least_one_is_not_none


class ToolkitSequenceRowsWrite(CogniteResource):
    """An object representing a list of rows from a sequence.

    Args:
        rows (typing.Sequence[SequenceRow]): The sequence rows.
        columns (collections.abc.Sequence[str): The column information.
        id (int | None): Identifier of the sequence the data belong to
        external_id (str | None): External id of the sequence the data belong to
    """

    def __init__(
        self,
        rows: collections.abc.Sequence[SequenceRow],
        columns: collections.abc.Sequence[str],
        id: int | None = None,
        external_id: str | None = None,
    ) -> None:
        if not at_least_one_is_not_none(id, external_id):
            raise ValueError("At least one of id and external_id must be specified.")

        col_length = len(columns)
        if wrong_length := [r for r in rows if len(r.values) != col_length]:
            raise ValueError(
                f"Rows {[r.row_number for r in wrong_length]} have wrong number of values, expected {col_length}"
            )
        self.rows = rows
        self.columns = columns
        self.id = id
        self.external_id = external_id

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        """Dump the sequence data into a json serializable Python data type.

        Args:
            camel_case (bool): Use camelCase for attribute names. Defaults to True.

        Returns:
            dict[str, Any]: A dictionary representing the instance.
        """
        key = "rowNumber" if camel_case else "row_number"
        dumped: dict[str, Any] = {
            "columns": list(self.columns),
            "rows": [{key: r.row_number, "values": r.values} for r in self.rows],
        }
        if self.id is not None:
            dumped["id"] = self.id
        if self.external_id is not None:
            dumped["externalId" if camel_case else "external_id"] = self.external_id
        return dumped

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> "ToolkitSequenceRowsWrite":
        return cls(
            rows=[SequenceRow._load(r) for r in resource["rows"]],
            columns=resource["columns"],
            id=resource.get("id"),
            external_id=resource.get("externalId"),
        )

    def as_sequence_rows(self) -> SequenceRows:
        return SequenceRows(
            self.rows,
            SequenceColumnList([SequenceColumn(external_id=ext_id) for ext_id in self.columns]),
            self.id,
            self.external_id,
        )


class ToolkitSequenceRows(WriteableCogniteResource[ToolkitSequenceRowsWrite]):
    """An object representing a list of rows from a sequence.

    Args:
        rows (typing.Sequence[SequenceRow]): The sequence rows.
        columns (SequenceColumnList): The column information.
        id (int | None): Identifier of the sequence the data belong to
        external_id (str | None): External id of the sequence the data belong to
    """

    def __init__(
        self,
        rows: collections.abc.Sequence[SequenceRow],
        columns: SequenceColumnList,
        id: int,
        external_id: str | None = None,
    ) -> None:
        col_length = len(columns)
        if wrong_length := [r for r in rows if len(r.values) != col_length]:
            raise ValueError(
                f"Rows {[r.row_number for r in wrong_length]} have wrong number of values, expected {col_length}"
            )
        self.rows = rows
        self.columns = columns
        self.id = id
        self.external_id = external_id

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        """Dump the sequence data into a json serializable Python data type.

        Args:
            camel_case (bool): Use camelCase for attribute names. Defaults to True.

        Returns:
            dict[str, Any]: A dictionary representing the instance.
        """
        key = "rowNumber" if camel_case else "row_number"
        dumped: dict[str, Any] = {
            "columns": self.columns.dump(camel_case),
            "rows": [{key: r.row_number, "values": r.values} for r in self.rows],
        }
        if self.id is not None:
            dumped["id"] = self.id
        if self.external_id is not None:
            dumped["externalId" if camel_case else "external_id"] = self.external_id
        return dumped

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> "ToolkitSequenceRows":
        return cls(
            rows=[SequenceRow._load(r) for r in resource["rows"]],
            columns=SequenceColumnList._load(resource["columns"]),
            id=resource["id"],
            external_id=resource.get("externalId"),
        )

    def as_write(self) -> ToolkitSequenceRowsWrite:
        return ToolkitSequenceRowsWrite(self.rows, [c.external_id for c in self.columns], self.id, self.external_id)  # type: ignore[misc]


class ToolkitSequenceRowsWriteList(CogniteResourceList, IdTransformerMixin):
    _RESOURCE = ToolkitSequenceRowsWrite


class ToolkitSequenceRowsList(
    WriteableCogniteResourceList[ToolkitSequenceRowsWrite, ToolkitSequenceRows], IdTransformerMixin
):
    _RESOURCE = ToolkitSequenceRows

    def as_write(self) -> ToolkitSequenceRowsWriteList:
        return ToolkitSequenceRowsWriteList([item.as_write() for item in self])



from collections.abc import Iterable

import pytest

from cognite_toolkit._cdf_tk.resource_classes.containers import ContainerYAML
from tests.test_unit.utils import find_resources


def container_property_type_cases() -> Iterable:
    container_property_types = [
        {
            "TextProperty": {
                "space": "my_space",
                "externalId": "my_container",
                "name": "My Container",
                "description": "A test container with text property",
                "properties": {
                    "textField": {
                        "name": "Text Field",
                        "description": "A simple text field",
                        "type": {
                            "type": "text",
                            "collation": "ucs_basic",
                            "maxListSize": 10,
                        },
                    }
                },
            }
        },
        {
            "BooleanProperty": {
                "space": "my_space",
                "externalId": "bool_container",
                "properties": {"isActive": {"type": {"type": "boolean"}}},
            }
        },
        {
            "FloatProperty": {
                "space": "my_space",
                "externalId": "float_container",
                "properties": {
                    "temperature": {
                        "type": {"type": "float32", "unit": {"externalId": "temperature", "sourceUnit": "celsius"}}
                    }
                },
            }
        },
        {
            "DirectRelationProperty": {
                "space": "my_space",
                "externalId": "relation_container",
                "properties": {
                    "connectedTo": {
                        "type": {
                            "type": "direct",
                            "container": {"type": "container", "space": "my_space", "externalId": "target_container"},
                        }
                    }
                },
            }
        },
        {
            "EnumProperty": {
                "space": "my_space",
                "externalId": "enum_container",
                "properties": {
                    "status": {
                        "type": {
                            "type": "enum",
                            "values": {
                                "active": {"name": "Active", "description": "Resource is active"},
                                "inactive": {"name": "Inactive", "description": "Resource is inactive"},
                            },
                        }
                    }
                },
            }
        },
    ]
    yield from (pytest.param(next(iter(pt.values())), id=next(iter(pt.keys()))) for pt in container_property_types)


def container_with_constraints_cases() -> Iterable:
    container_constraints = [
        {
            "UniquenessConstraint": {
                "space": "my_space",
                "externalId": "unique_container",
                "properties": {"identifier": {"type": {"type": "text"}}},
                "constraints": {
                    "uniqueIdentifier": {"constraintType": "uniqueness", "properties": ["identifier"], "bySpace": True}
                },
            }
        },
        {
            "RequiresConstraint": {
                "space": "my_space",
                "externalId": "requires_container",
                "properties": {"name": {"type": {"type": "text"}}},
                "constraints": {
                    "requiresOtherContainer": {
                        "constraintType": "requires",
                        "require": {"type": "container", "space": "my_space", "externalId": "dependency_container"},
                    }
                },
            }
        },
    ]
    yield from (pytest.param(next(iter(c.values())), id=next(iter(c.keys()))) for c in container_constraints)


def container_with_indexes_cases() -> Iterable:
    container_indexes = [
        {
            "BtreeIndex": {
                "space": "my_space",
                "externalId": "indexed_container",
                "properties": {"timestamp": {"type": {"type": "timestamp"}}, "value": {"type": {"type": "float64"}}},
                "indexes": {
                    "timeSeriesIndex": {
                        "indexType": "btree",
                        "properties": ["timestamp", "value"],
                        "bySpace": True,
                        "cursorable": True,
                    }
                },
            }
        }
    ]
    yield from (pytest.param(next(iter(idx.values())), id=next(iter(idx.keys()))) for idx in container_indexes)


class TestContainerYAML:
    @pytest.mark.parametrize("data", list(find_resources("Container")))
    def test_load_valid_container_file(self, data: dict[str, object]) -> None:
        loaded = ContainerYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data", container_property_type_cases())
    def test_load_valid_container_property_types(self, data: dict[str, object]) -> None:
        loaded = ContainerYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data", container_with_constraints_cases())
    def test_load_valid_container_constraints(self, data: dict[str, object]) -> None:
        loaded = ContainerYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data", container_with_indexes_cases())
    def test_load_valid_container_indexes(self, data: dict[str, object]) -> None:
        loaded = ContainerYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    def test_invalid_external_id(self) -> None:
        invalid_data = {
            "space": "my_space",
            "externalId": "Query",
            "properties": {"field": {"type": {"type": "text"}}},
        }

        with pytest.raises(ValueError, match="is a reserved container External ID"):
            ContainerYAML.model_validate(invalid_data)

    def test_invalid_property_identifier(self) -> None:
        invalid_data = {
            "space": "my_space",
            "externalId": "valid_container",
            "properties": {"invalid@": {"type": {"type": "text"}}},
        }

        with pytest.raises(ValueError, match="does not match the required pattern"):
            ContainerYAML.model_validate(invalid_data)

    def test_forbidden_property_identifier(self) -> None:
        invalid_data = {
            "space": "my_space",
            "externalId": "valid_container",
            "properties": {"edge_id": {"type": {"type": "text"}}},
        }

        with pytest.raises(ValueError, match="is a reserved property identifier"):
            ContainerYAML.model_validate(invalid_data)

    def test_space_format_validation(self) -> None:
        invalid_data = {
            "space": "invalid space",  # Contains a space
            "externalId": "valid_container",
            "properties": {"field": {"type": {"type": "text"}}},
        }

        with pytest.raises(ValueError):
            ContainerYAML.model_validate(invalid_data)



from collections.abc import Callable, Hashable
from typing import Any


def diff_list_hashable(local: list[Hashable], cdf: list[Hashable]) -> tuple[dict[int, int], list[int]]:
    local_by_cdf: dict[int, int] = {}
    added: list[int] = []
    index_by_local = {item: i for i, item in enumerate(local)}
    for index, item in enumerate(cdf):
        if item in index_by_local:
            local_by_cdf[index_by_local[item]] = index
        else:
            added.append(index)
    return local_by_cdf, added


def diff_list_identifiable(
    local: list[Any], cdf: list[Any], *, get_identifier: Callable[[Any], Hashable]
) -> tuple[dict[int, int], list[int]]:
    return diff_list_hashable([get_identifier(item) for item in local], [get_identifier(item) for item in cdf])


def diff_list_force_hashable(local: Any, cdf: Any) -> tuple[dict[int, int], list[int]]:
    return diff_list_identifiable(local, cdf, get_identifier=force_hash)


def force_hash(item: Any) -> int:
    if isinstance(item, dict):
        return hash_dict(item)
    if isinstance(item, list):
        return hash_list(item)
    if isinstance(item, Hashable):
        return hash(item)
    raise ValueError(f"Cannot hash value {item}")


def hash_dict(d: dict) -> int:
    hash_ = 0
    for key, value in sorted(d.items(), key=lambda x: x[0]):
        if isinstance(value, dict):
            hash_ ^= hash_dict(value)
        elif isinstance(value, list):
            hash_ ^= hash_list(value)
        elif isinstance(value, Hashable):
            hash_ ^= hash((key, value))
        else:
            raise ValueError(f"Cannot hash value {value}")
    return hash_


def hash_list(lst: list) -> int:
    hash_ = 0
    for i, item in enumerate(lst):
        if isinstance(item, dict):
            hash_ ^= hash_dict(item)
        elif isinstance(item, list):
            hash_ ^= hash_list(item)
        elif isinstance(item, Hashable):
            hash_ ^= hash((i, item))
        else:
            raise ValueError(f"Cannot hash value {item}")
    return hash_


def dm_identifier(data: dict[str, Any]) -> tuple[str, ...]:
    return data.get("type", ""), data["space"], data["externalId"], data.get("version", "")



import queue
import threading
from collections.abc import Callable, Iterable, Sized
from typing import Generic, TypeVar

from rich.console import Console
from rich.progress import BarColumn, Progress, TaskID, TaskProgressColumn, TextColumn, TimeRemainingColumn

T_Download = TypeVar("T_Download", bound=Sized)
T_Processed = TypeVar("T_Processed", bound=Sized)


class ProducerWorkerExecutor(Generic[T_Download, T_Processed]):
    """This class manages the execution of a producer-worker pattern with multiple threads.

    It downloads data from an iterable, processes it, and writes the processed data.

    Args:
        download_iterable (Iterable[T_Download]): An iterable that yields chunks of items.
        process (Callable[[T_Download], T_Processed]): A function that processes the chunks of items.
        write (Callable[[T_Processed], None]): A function that writes the processed items. This is typically to
            a file, CDF, or another storage format.
        iteration_count (int): The total number of iterations expected. Note this is not the total number of items,
            but rather the number of chunks of items that will be processed.
        max_queue_size (int): The maximum size of the queues used for processing and writing. Note that the queue
            holds chunks of items, not individual items, so this should be set based on the expected size of the
            chunks being processed.
        download_description (str): A description of the download process, used for progress tracking.
        process_description (str): A description of the processing step, used for progress tracking.
        write_description (str): A description of the writing step, used for progress tracking.
        console (Console | None): An optional Rich Console instance for outputting progress and error messages.

    Examples:
        >>> from cognite_toolkit._cdf_tk.utils.producer_worker import ProducerWorkerExecutor
        >>> from typing import List
        >>> def download_data() -> Iterable[list[int]]:
        ...     yield [1, 2, 3, 4, 5]
        ...     yield [6, 7, 8, 9, 10]
        >>> def process_data(data: list[int]) -> list[int]:
        ...     return [x * 2 for x in data]
        >>> def write_data(data: list[int]) -> None:
        ...     print(f"Writing data: {data}")
        >>> executor = ProducerWorkerExecutor(
        ...     download_iterable=download_data(),
        ...     process=process_data,
        ...     write=write_data,
        ...     iteration_count=2,
        ...     max_queue_size=1,
        ... )
        >>> executor.run()

    """

    def __init__(
        self,
        download_iterable: Iterable[T_Download],
        process: Callable[[T_Download], T_Processed],
        write: Callable[[T_Processed], None],
        iteration_count: int,
        max_queue_size: int,
        download_description: str = "downloading",
        process_description: str = "processing",
        write_description: str = "writing to file",
        console: Console | None = None,
    ) -> None:
        self._download_iterable = download_iterable
        self._process = process
        self._write = write
        self.iteration_count = iteration_count
        self.download_description = download_description
        self.process_description = process_description
        self.write_description = write_description
        self.console = console or Console()

        self.download_complete = False
        self.is_processing = False

        # Queues for managing the flow of data between threads
        # Download -> [process_queue] -> Process -> [write_queue] -> Write
        self.process_queue: queue.Queue[T_Download] = queue.Queue(maxsize=max_queue_size)
        self.write_queue: queue.Queue[T_Processed] = queue.Queue(maxsize=max_queue_size)

        self.total_items = 0
        self.error_occurred = False
        self.error_message = ""

    def run(self) -> None:
        with Progress(
            TextColumn("[bold blue]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
            TimeRemainingColumn(),
            console=self.console,
        ) as progress:
            download_task = progress.add_task(self.download_description.title(), total=self.iteration_count)
            download_thread = threading.Thread(target=self._download_worker, args=(progress, download_task))
            process_task = progress.add_task(self.process_description.title(), total=self.iteration_count)
            process_thread = threading.Thread(target=self._process_worker, args=(progress, process_task))

            write_task = progress.add_task(self.download_description.title(), total=self.iteration_count)
            write_thread = threading.Thread(target=self._write_worker, args=(progress, write_task))

            download_thread.start()
            process_thread.start()
            write_thread.start()

            # Wait for all threads to finish
            download_thread.join()
            process_thread.join()
            write_thread.join()

    def _download_worker(self, progress: Progress, download_task: TaskID) -> None:
        """Worker thread for downloading data."""
        iterator = iter(self._download_iterable)
        while not self.error_occurred:
            try:
                items = next(iterator)
                self.total_items += len(items)
                while not self.error_occurred:
                    try:
                        self.process_queue.put(items, timeout=0.5)
                        progress.update(download_task, advance=1)
                        break  # Exit the loop once the item is successfully added
                    except queue.Full:
                        # Retry until the queue has space
                        continue
            except StopIteration:
                self.download_complete = True
                break
            except Exception as e:
                self.error_occurred = True
                self.error_message = str(e)
                self.console.print(f"[red]Error[/red] occurred while {self.download_description}: {self.error_message}")
                break

    def _process_worker(self, progress: Progress, process_task: TaskID) -> None:
        """Worker thread for processing data."""
        self.is_processing = True
        while (not self.download_complete or not self.process_queue.empty()) and not self.error_occurred:
            try:
                items = self.process_queue.get(timeout=0.5)
                processed_items = self._process(items)
                self.write_queue.put(processed_items)
                progress.update(process_task, advance=1)
                self.process_queue.task_done()
            except queue.Empty:
                continue
            except Exception as e:
                self.error_occurred = True
                self.error_message = str(e)
                self.console.print(
                    f"[red]ErrorError[/red] occurred while {self.process_description}: {self.error_message}"
                )
                break
        self.is_processing = False

    def _write_worker(self, progress: Progress, write_task: TaskID) -> None:
        """Worker thread for writing data to file."""
        while (
            not self.download_complete
            or self.is_processing
            or not self.write_queue.empty()
            or not self.process_queue.empty()
        ) and not self.error_occurred:
            try:
                items = self.write_queue.get(timeout=0.5)
                self._write(items)
                progress.update(write_task, advance=1)
                self.write_queue.task_done()
            except queue.Empty:
                continue
            except Exception as e:
                self.error_occurred = True
                self.error_message = str(e)
                self.console.print(f"[red]Error[/red] occurred while {self.write_description}: {self.error_message}")
                break



# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import re
from collections.abc import Hashable, Iterable, Sequence
from functools import lru_cache
from pathlib import Path
from typing import Any, final

import yaml
from cognite.client.data_classes import (
    ExtractionPipeline,
    ExtractionPipelineConfig,
    ExtractionPipelineList,
)
from cognite.client.data_classes.capabilities import (
    Capability,
    ExtractionConfigsAcl,
    ExtractionPipelinesAcl,
)
from cognite.client.data_classes.extractionpipelines import (
    ExtractionPipelineConfigList,
    ExtractionPipelineConfigWrite,
    ExtractionPipelineConfigWriteList,
    ExtractionPipelineWrite,
    ExtractionPipelineWriteList,
)
from cognite.client.exceptions import CogniteAPIError, CogniteDuplicatedError, CogniteNotFoundError
from cognite.client.utils.useful_types import SequenceNotStr
from rich import print

from cognite_toolkit._cdf_tk._parameters import ANYTHING, ParameterSpec, ParameterSpecSet
from cognite_toolkit._cdf_tk.client.data_classes.raw import RawDatabase, RawTable
from cognite_toolkit._cdf_tk.constants import BUILD_FOLDER_ENCODING
from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitRequiredValueError,
)
from cognite_toolkit._cdf_tk.loaders._base_loaders import ResourceLoader
from cognite_toolkit._cdf_tk.tk_warnings import (
    HighSeverityWarning,
)
from cognite_toolkit._cdf_tk.utils import (
    load_yaml_inject_variables,
    read_yaml_content,
    safe_read,
    stringify_value_by_key_in_yaml,
)
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_force_hashable, diff_list_identifiable

from .auth_loaders import GroupAllScopedLoader
from .data_organization_loaders import DataSetsLoader
from .raw_loaders import RawDatabaseLoader, RawTableLoader


@final
class ExtractionPipelineLoader(
    ResourceLoader[
        str, ExtractionPipelineWrite, ExtractionPipeline, ExtractionPipelineWriteList, ExtractionPipelineList
    ]
):
    folder_name = "extraction_pipelines"
    filename_pattern = r"^(?:(?!\.config).)*$"  # Matches all yaml files except file names who's stem contain *.config.
    resource_cls = ExtractionPipeline
    resource_write_cls = ExtractionPipelineWrite
    list_cls = ExtractionPipelineList
    list_write_cls = ExtractionPipelineWriteList
    kind = "ExtractionPipeline"
    dependencies = frozenset({DataSetsLoader, RawDatabaseLoader, RawTableLoader, GroupAllScopedLoader})
    _doc_url = "Extraction-Pipelines/operation/createExtPipes"

    @property
    def display_name(self) -> str:
        return "extraction pipelines"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[ExtractionPipelineWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [ExtractionPipelinesAcl.Action.Read]
            if read_only
            else [ExtractionPipelinesAcl.Action.Read, ExtractionPipelinesAcl.Action.Write]
        )

        scope: ExtractionPipelinesAcl.Scope.All | ExtractionPipelinesAcl.Scope.DataSet = (  # type: ignore[valid-type]
            ExtractionPipelinesAcl.Scope.All()
        )
        if items is not None:
            if data_set_id := {item.data_set_id for item in items if item.data_set_id}:
                scope = ExtractionPipelinesAcl.Scope.DataSet(list(data_set_id))

        return ExtractionPipelinesAcl(
            actions,
            scope,  # type: ignore[arg-type]
        )

    @classmethod
    def get_id(cls, item: ExtractionPipeline | ExtractionPipelineWrite | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if item.external_id is None:
            raise ToolkitRequiredValueError("ExtractionPipeline must have external_id set.")
        return item.external_id

    @classmethod
    def get_internal_id(cls, item: ExtractionPipeline | dict) -> int:
        if isinstance(item, dict):
            return item["id"]
        if item.id is None:
            raise ToolkitRequiredValueError("ExtractionPipeline must have id set.")
        return item.id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceLoader], Hashable]]:
        seen_databases: set[str] = set()
        if "dataSetExternalId" in item:
            yield DataSetsLoader, item["dataSetExternalId"]
        if "rawTables" in item:
            for entry in item["rawTables"]:
                if db := entry.get("dbName"):
                    if db not in seen_databases:
                        seen_databases.add(db)
                        yield RawDatabaseLoader, RawDatabase(db_name=db)
                    if "tableName" in entry:
                        yield RawTableLoader, RawTable._load(entry)

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> ExtractionPipelineWrite:
        if ds_external_id := resource.pop("dataSetExternalId", None):
            resource["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        if "createdBy" not in resource:
            # Todo; Bug SDK missing default value (this will be set on the server-side if missing)
            resource["createdBy"] = "unknown"
        return ExtractionPipelineWrite._load(resource)

    def dump_resource(self, resource: ExtractionPipeline, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        local = local or {}
        if data_set_id := dumped.pop("dataSetId", None):
            dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
        if dumped.get("createdBy") == "unknown" and "createdBy" not in local:
            dumped.pop("createdBy", None)
        return dumped

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path == ("rawTables",):
            return diff_list_identifiable(local, cdf, get_identifier=lambda x: (x["dbName"], x["tableName"]))
        return super().diff_list(local, cdf, json_path)

    def create(self, items: Sequence[ExtractionPipelineWrite]) -> ExtractionPipelineList:
        items = list(items)
        try:
            return self.client.extraction_pipelines.create(items)
        except CogniteDuplicatedError as e:
            if len(e.duplicated) < len(items):
                for dup in e.duplicated:
                    ext_id = dup.get("externalId", None)
                    for item in items:
                        if item.external_id == ext_id:
                            items.remove(item)

                return self.client.extraction_pipelines.create(items)
        return ExtractionPipelineList([])

    def retrieve(self, ids: SequenceNotStr[str]) -> ExtractionPipelineList:
        return self.client.extraction_pipelines.retrieve_multiple(external_ids=ids, ignore_unknown_ids=True)

    def update(self, items: ExtractionPipelineWriteList) -> ExtractionPipelineList:
        # Bug in SDK overload so need the ignore.
        return self.client.extraction_pipelines.update(items, mode="replace")  # type: ignore[call-overload]

    def delete(self, ids: SequenceNotStr[str | int]) -> int:
        internal_ids, external_ids = self._split_ids(ids)
        try:
            self.client.extraction_pipelines.delete(id=internal_ids, external_id=external_ids)
        except CogniteNotFoundError as e:
            not_existing = {external_id for dup in e.not_found if (external_id := dup.get("externalId", None))}
            if id_list := [id_ for id_ in ids if id_ not in not_existing]:
                internal_ids, external_ids = self._split_ids(id_list)
                self.client.extraction_pipelines.delete(id=internal_ids, external_id=external_ids)
        except CogniteAPIError as e:
            if e.code == 403 and "not found" in e.message and "extraction pipeline" in e.message.lower():
                return 0
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[ExtractionPipeline]:
        if data_set_external_id is None:
            yield from iter(self.client.extraction_pipelines)
            return
        data_set = self.client.data_sets.retrieve(external_id=data_set_external_id)
        if data_set is None:
            raise ToolkitRequiredValueError(f"DataSet {data_set_external_id!r} does not exist")
        for pipeline in self.client.extraction_pipelines:
            if pipeline.data_set_id == data_set.id:
                yield pipeline

    @classmethod
    @lru_cache(maxsize=1)
    def get_write_cls_parameter_spec(cls) -> ParameterSpecSet:
        spec = super().get_write_cls_parameter_spec()
        # Added by toolkit
        spec.add(ParameterSpec(("dataSetExternalId",), frozenset({"str"}), is_required=False, _is_nullable=False))
        # Set on deploy time by toolkit
        spec.discard(ParameterSpec(("dataSetId",), frozenset({"int"}), is_required=True, _is_nullable=False))
        return spec


@final
class ExtractionPipelineConfigLoader(
    ResourceLoader[
        str,
        ExtractionPipelineConfigWrite,
        ExtractionPipelineConfig,
        ExtractionPipelineConfigWriteList,
        ExtractionPipelineConfigList,
    ]
):
    folder_name = "extraction_pipelines"
    filename_pattern = r"^.*config$"
    resource_cls = ExtractionPipelineConfig
    resource_write_cls = ExtractionPipelineConfigWrite
    list_cls = ExtractionPipelineConfigList
    list_write_cls = ExtractionPipelineConfigWriteList
    kind = "Config"
    dependencies = frozenset({ExtractionPipelineLoader})
    _doc_url = "Extraction-Pipelines-Config/operation/createExtPipeConfig"

    @property
    def display_name(self) -> str:
        return "extraction pipeline configs"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[ExtractionPipelineConfigWrite] | None, read_only: bool
    ) -> list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [ExtractionConfigsAcl.Action.Read]
            if read_only
            else [ExtractionConfigsAcl.Action.Read, ExtractionConfigsAcl.Action.Write]
        )

        return [
            ExtractionConfigsAcl(
                actions,
                ExtractionConfigsAcl.Scope.All(),
            )
        ]

    @classmethod
    def get_id(cls, item: ExtractionPipelineConfig | ExtractionPipelineConfigWrite | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if item.external_id is None:
            raise ToolkitRequiredValueError("ExtractionPipelineConfig must have external_id set.")
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceLoader], Hashable]]:
        if "externalId" in item:
            yield ExtractionPipelineLoader, item["externalId"]

    def safe_read(self, filepath: Path | str) -> str:
        # The config is expected to be a string that is parsed as a YAML on the server side.
        # The user typically writes the config as an object, so add a | to ensure it is parsed as a string.
        return stringify_value_by_key_in_yaml(safe_read(filepath, encoding=BUILD_FOLDER_ENCODING), key="config")

    def load_resource_file(
        # special case where the environment variable keys in the 'config' value
        # should not be replaced but preserved as is
        self,
        filepath: Path,
        environment_variables: dict[str, str | None] | None = None,
    ) -> list[dict[str, Any]]:
        raw_str = self.safe_read(filepath)

        original = load_yaml_inject_variables(raw_str, {}, validate=False, original_filepath=filepath)
        replaced = load_yaml_inject_variables(
            raw_str, environment_variables or {}, validate=False, original_filepath=filepath
        )

        if isinstance(original, dict) and isinstance(replaced, dict):
            if "config" in original:
                replaced["config"] = original.get("config")
                return [replaced]

        for orig_item, repl_item in zip(original, replaced):
            if "config" in orig_item:
                repl_item["config"] = orig_item.get("config")  # type: ignore
        return replaced if isinstance(replaced, list) else [replaced]

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> ExtractionPipelineConfigWrite:
        config_raw = resource.get("config")
        if isinstance(config_raw, str):
            # There might be keyvauls secrets in the config that would lead to parsing errors. The syntax
            # for this is `connection-string: !keyvault secret`. This is not valid YAML, so we need to
            # replace it with `connection-string: keyvault secret` to make it valid.
            config_raw = re.sub(r": !(\w+)", r": \1", config_raw)
            try:
                read_yaml_content(config_raw)
            except yaml.YAMLError as e:
                print(
                    HighSeverityWarning(
                        f"Configuration for {resource.get('external_id', 'missing')} could not be parsed "
                        f"as valid YAML, which is the recommended format. Error: {e}"
                    ).get_message()
                )
        return ExtractionPipelineConfigWrite._load(resource)

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path[0] == "config":
            # Assume all arrays in the config are hashable
            return diff_list_force_hashable(local, cdf)
        return super().diff_list(local, cdf, json_path)

    def _upsert(self, items: ExtractionPipelineConfigWriteList) -> ExtractionPipelineConfigList:
        upserted = ExtractionPipelineConfigList([])
        for item in items:
            created = self.client.extraction_pipelines.config.create(item)
            upserted.append(created)
        return upserted

    def create(self, items: ExtractionPipelineConfigWriteList) -> ExtractionPipelineConfigList:
        return self._upsert(items)

    # configs cannot be updated, instead new revision is created
    def update(self, items: ExtractionPipelineConfigWriteList) -> ExtractionPipelineConfigList:
        return self._upsert(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> ExtractionPipelineConfigList:
        retrieved = ExtractionPipelineConfigList([])
        for external_id in ids:
            try:
                config_retrieved = self.client.extraction_pipelines.config.retrieve(external_id=external_id)
            except CogniteAPIError as e:
                if (
                    e.code == 404
                    and e.message.startswith("There is no config stored for the extraction pipeline with external id")
                ) or e.message.startswith("Extraction pipeline not found"):
                    continue
                raise e
            if config_retrieved:
                retrieved.append(config_retrieved)
        return retrieved

    def delete(self, ids: SequenceNotStr[str]) -> int:
        """Delete is not supported for extraction pipeline configs.

        Instead, we assume that when the user deletes the extraction pipeline configs, they are also deleting the
        extraction pipelines which will automatically delete the configs. In this method, we simply count the number
        of configs that exist for the given ids and return that number as these will be deleted.
        """
        count = 0
        for id_ in ids:
            try:
                result = self.client.extraction_pipelines.config.list(external_id=id_)
            except CogniteAPIError as e:
                if e.code == 403 and "not found" in e.message and "extraction pipeline" in e.message.lower():
                    continue
            else:
                if result:
                    count += 1
        return count

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[ExtractionPipelineConfig]:
        parent_iterable = parent_ids or iter(self.client.extraction_pipelines)
        for parent_id in parent_iterable or []:
            pipeline_id: str | None = None
            if isinstance(parent_id, ExtractionPipeline):
                if parent_id.external_id:
                    pipeline_id = parent_id.external_id
            elif isinstance(parent_id, str):
                pipeline_id = parent_id

            if pipeline_id is None:
                continue

            try:
                yield self.client.extraction_pipelines.config.retrieve(external_id=pipeline_id)
            except CogniteAPIError as e:
                if e.code == 404 and "There is no config stored" in e.message:
                    continue

    @classmethod
    @lru_cache(maxsize=1)
    def get_write_cls_parameter_spec(cls) -> ParameterSpecSet:
        spec = super().get_write_cls_parameter_spec()
        # Added by toolkit
        spec.add(ParameterSpec(("config", ANYTHING), frozenset({"dict"}), is_required=True, _is_nullable=False))
        return spec



from .base import (
    GeneralWarning,
    SeverityLevel,
    ToolkitWarning,
    WarningList,
    catch_warnings,
)
from .fileread import (
    CaseTypoWarning,
    DataSetMissingWarning,
    DuplicatedItemWarning,
    EnvironmentVariableMissingWarning,
    FileExistsWarning,
    FileReadWarning,
    MissingFileWarning,
    MissingReferencedWarning,
    MissingRequiredParameterWarning,
    NamespacingConventionWarning,
    NamingConventionWarning,
    PrefixConventionWarning,
    ResourceMissingIdentifierWarning,
    SourceFileModifiedWarning,
    StreamlitRequirementsWarning,
    TemplateVariableWarning,
    UnresolvedVariableWarning,
    UnusedParameterWarning,
    YAMLFileWarning,
    YAMLFileWithElementWarning,
)
from .other import (
    HighSeverityWarning,
    HTTPWarning,
    IgnoredValueWarning,
    IncorrectResourceWarning,
    LowSeverityWarning,
    MediumSeverityWarning,
    MissingCapabilityWarning,
    MissingDependencyWarning,
    ToolkitBugWarning,
    ToolkitDependenciesIncludedWarning,
    ToolkitDeprecationWarning,
    ToolkitNotSupportedWarning,
    UnexpectedFileLocationWarning,
)

__all__ = [
    "CaseTypoWarning",
    "DataSetMissingWarning",
    "DuplicatedItemWarning",
    "EnvironmentVariableMissingWarning",
    "FileExistsWarning",
    "FileReadWarning",
    "GeneralWarning",
    "HTTPWarning",
    "HighSeverityWarning",
    "IgnoredValueWarning",
    "IncorrectResourceWarning",
    "LowSeverityWarning",
    "MediumSeverityWarning",
    "MissingCapabilityWarning",
    "MissingDependencyWarning",
    "MissingFileWarning",
    "MissingReferencedWarning",
    "MissingRequiredParameterWarning",
    "NamespacingConventionWarning",
    "NamingConventionWarning",
    "PrefixConventionWarning",
    "ResourceMissingIdentifierWarning",
    "SeverityLevel",
    "SourceFileModifiedWarning",
    "StreamlitRequirementsWarning",
    "TemplateVariableWarning",
    "ToolkitBugWarning",
    "ToolkitDependenciesIncludedWarning",
    "ToolkitDeprecationWarning",
    "ToolkitNotSupportedWarning",
    "ToolkitWarning",
    "UnexpectedFileLocationWarning",
    "UnresolvedVariableWarning",
    "UnusedParameterWarning",
    "WarningList",
    "YAMLFileWarning",
    "YAMLFileWithElementWarning",
    "catch_warnings",
]



from __future__ import annotations

import shutil
import sys
import venv
from collections.abc import Mapping
from pathlib import Path
from subprocess import DEVNULL, Popen
from types import SimpleNamespace

from cognite_toolkit._cdf_tk.exceptions import ToolkitEnvError


class FunctionVirtualEnvironment(venv.EnvBuilder):
    def __init__(self, requirements_txt: Path | str, rebuild: bool) -> None:
        super().__init__(
            system_site_packages=False,
            clear=rebuild,
            with_pip=True,
        )
        self.requirements_txt = requirements_txt
        self._context: SimpleNamespace | None = None

    def post_setup(self, context: SimpleNamespace) -> None:
        args = [str(Path(context.bin_path) / "pip"), "install", "--disable-pip-version-check", "-r", "requirements.txt"]

        function_dir = Path(context.env_dir).parent
        requirements_destination_path = function_dir / "requirements.txt"
        if isinstance(self.requirements_txt, Path):
            shutil.copy(self.requirements_txt, requirements_destination_path)
        else:
            requirements_destination_path.write_text(self.requirements_txt, encoding="utf-8")

        process = Popen(args, stdout=DEVNULL, stderr=sys.stderr, cwd=str(function_dir))
        process.wait()
        if process.returncode != 0:
            suffix = f" {self.requirements_txt.as_posix()}" if isinstance(self.requirements_txt, Path) else ""
            raise ToolkitEnvError(
                f"Unable to install dependencies in 'requirements.txt' in the current environment {suffix}."
            )
        self._context = context

    def execute(self, script: Path, script_name: str, env: Mapping[str, str] | None = None) -> None:
        if self._context is None:
            raise ToolkitEnvError("Virtual environment not created.")
        function_dir = Path(self._context.env_dir).parent
        args = [str(Path(self._context.bin_path) / "python"), str(script)]

        process = Popen(args, stdout=sys.stdout, stderr=sys.stderr, cwd=str(function_dir), env=env)
        process.wait()
        if process.returncode != 0:
            raise ToolkitEnvError(f"Error executing {script_name} {script.as_posix()}.")



import io

from PIL import Image


def batch(iterable, n=1):
    """Batch iterable."""
    len_iterable = len(iterable)
    for ndx in range(0, len_iterable, n):
        yield iterable[ndx : min(ndx + n, len_iterable)]


def image_to_byte_array(image: Image) -> bytes:
    """Convert PIL image to byte array."""
    img_byte_arr = io.BytesIO()
    image.save(img_byte_arr, format=image.format)
    return img_byte_arr.getvalue()



from collections.abc import Callable, Iterable, Sequence
from graphlib import CycleError, TopologicalSorter
from typing import Any

from cognite_toolkit._cdf_tk.builders._base import Builder
from cognite_toolkit._cdf_tk.data_classes._build_files import BuildDestinationFile, BuildSourceFile
from cognite_toolkit._cdf_tk.data_classes._module_directories import ModuleLocation
from cognite_toolkit._cdf_tk.exceptions import ToolkitError
from cognite_toolkit._cdf_tk.loaders._resource_loaders.location_loaders import LocationFilterLoader
from cognite_toolkit._cdf_tk.tk_warnings.base import ToolkitWarning, WarningList
from cognite_toolkit._cdf_tk.tk_warnings.fileread import FileReadWarning


class LocationBuilder(Builder):
    _resource_folder = LocationFilterLoader.folder_name

    def build(
        self, source_files: list[BuildSourceFile], module: ModuleLocation, console: Callable[[str], None] | None = None
    ) -> Iterable[BuildDestinationFile | Sequence[ToolkitWarning]]:
        location_by_external_id: dict[str, tuple[dict[str, Any], BuildSourceFile]] = {}
        location_hierarchy_graph: dict[str, list[Any]] = {}

        # Ordering all location filters in to ensure correct hierarchy dependency
        # within the module. This is required by the Location API.
        # Doing this in three stages:
        # 1. collect all locations across source files,
        # 2. sort them in a topological order,
        # 3. create a new file for each location where the prefix index ensures deployment order
        # ... while also maintaining reference to source file

        for source_file in source_files:
            loader, warning = self._get_loader(source_file.source.path)
            if isinstance(loader, LocationFilterLoader):
                if warning is not None:
                    yield [warning]
                continue

            loaded_locations = (
                source_file.loaded
                if isinstance(source_file.loaded, list)
                else [source_file.loaded]
                if source_file.loaded
                else []
            )
            for loaded_location in loaded_locations:
                ext_id = loaded_location.get("externalId")
                parent_external_id = loaded_location.get("parentExternalId")

                if ext_id:
                    location_by_external_id[ext_id] = (loaded_location, source_file)
                    location_hierarchy_graph.setdefault(ext_id, [])

                    if parent_external_id:
                        location_hierarchy_graph.setdefault(parent_external_id, [])
                        location_hierarchy_graph[ext_id].append(parent_external_id)

            warnings = WarningList[FileReadWarning]()

        ordered_locations = []
        try:
            for external_id in TopologicalSorter(location_hierarchy_graph).static_order():
                (location, _) = location_by_external_id[external_id]
                ordered_locations.append(location)
        except CycleError:
            raise ToolkitError(
                "Circular dependency found in Locations. Locations must be hierarchical. Please check the externalId and parentExternalId fields."
            )

        for item in ordered_locations:
            external_id = item["externalId"]
            (location, build_source_file) = location_by_external_id[external_id]
            destination_path = self._create_destination_path(build_source_file.source.path, loader.kind)  # type: ignore[union-attr]

            yield BuildDestinationFile(
                path=destination_path,
                loaded=location,
                loader=loader,  # type: ignore[arg-type]
                source=build_source_file.source,
                extra_sources=None,
                warnings=warnings,
            )



from cognite.client import CogniteClient


def handle(data: dict, client: CogniteClient, secrets: dict, function_call_info: dict) -> dict:
    # This will fail unless the function has the specified capabilities.
    print("Print statements will be shown in the logs.")
    print("Running with the following configuration:\n")
    return {
        "data": data,
        "secrets": mask_secrets(secrets),
        "functionInfo": function_call_info,
    }


def mask_secrets(secrets: dict) -> dict:
    return {k: "***" for k in secrets}



from __future__ import annotations

import abc
import collections.abc
import dataclasses
import enum
import inspect
import random
import re
import string
import typing
from datetime import date, datetime
from pathlib import Path
from types import UnionType
from typing import IO, Any, Callable, Literal, Optional, TypeVar, get_args, get_origin

import pytest
from _pytest.monkeypatch import MonkeyPatch
from cognite.client import CogniteClient
from cognite.client._constants import MAX_VALID_INTERNAL_ID
from cognite.client.data_classes import (
    Datapoints,
    DatapointsArray,
    DataPointSubscriptionWrite,
    EndTimeFilter,
    FunctionTaskOutput,
    FunctionTaskParameters,
    Geometry,
    Relationship,
    SequenceColumn,
    SequenceColumnList,
    SequenceData,
    SequenceRow,
    SequenceRows,
    Transformation,
    TransformationScheduleWrite,
    filters,
)
from cognite.client.data_classes._base import CogniteResourceList
from cognite.client.data_classes.capabilities import Capability, LegacyCapability, UnknownAcl
from cognite.client.data_classes.data_modeling.data_types import ListablePropertyType
from cognite.client.data_classes.data_modeling.query import NodeResultSetExpression, Query
from cognite.client.data_classes.filters import Filter
from cognite.client.data_classes.hosted_extractors import (
    BodyLoad,
    HeaderValueLoad,
    NextUrlLoad,
    QueryParamLoad,
    RestConfig,
)
from cognite.client.data_classes.transformations.notifications import TransformationNotificationWrite
from cognite.client.data_classes.workflows import (
    WorkflowTaskOutput,
    WorkflowTaskParameters,
    WorkflowTriggerDataModelingQuery,
)
from cognite.client.testing import CogniteClientMock
from cognite.client.utils.useful_types import SequenceNotStr
from questionary import Choice

from cognite_toolkit._cdf_tk._parameters.get_type_hints import _TypeHints
from cognite_toolkit._cdf_tk.client.data_classes.location_filters import (
    LocationFilter,
    LocationFilterScene,
)
from cognite_toolkit._cdf_tk.client.data_classes.sequences import ToolkitSequenceRows
from cognite_toolkit._cdf_tk.constants import MODULES
from cognite_toolkit._cdf_tk.utils import load_yaml_inject_variables, read_yaml_file
from tests.data import COMPLETE_ORG

UNION_TYPES = {typing.Union, UnionType}


def mock_read_yaml_file(
    file_content_by_name: dict[str, dict | list], monkeypatch: MonkeyPatch, modify: bool = False
) -> None:
    def fake_read_yaml_file(
        filepath: Path, expected_output: Literal["list", "dict"] = "dict"
    ) -> dict[str, Any] | list[dict[str, Any]]:
        if file_content := file_content_by_name.get(filepath.name):
            if modify:
                source = read_yaml_file(filepath, expected_output)
                source.update(file_content)
                file_content = source
            return file_content
        return read_yaml_file(filepath, expected_output)

    def fake_load_yaml_inject_variables(
        filepath: Path | str,
        environment_variables: dict[str, str | None],
        required_return_type: Literal["any", "list", "dict"] = "any",
        validate: bool = True,
        original_filepath: Path | None = None,
    ) -> dict[str, Any] | list[dict[str, Any]]:
        if isinstance(filepath, str):
            return load_yaml_inject_variables(
                filepath, environment_variables, required_return_type, validate, original_filepath
            )
        if file_content := file_content_by_name.get(filepath.name):
            if modify:
                source = load_yaml_inject_variables(
                    filepath, environment_variables, required_return_type, validate, original_filepath
                )
                source.update(file_content)
                file_content = source
            return file_content
        return load_yaml_inject_variables(
            filepath, environment_variables, required_return_type, validate, original_filepath
        )

    monkeypatch.setattr("cognite_toolkit._cdf_tk.utils.read_yaml_file", fake_read_yaml_file)
    monkeypatch.setattr("cognite_toolkit._cdf_tk.data_classes._base.read_yaml_file", fake_read_yaml_file)
    monkeypatch.setattr("cognite_toolkit._cdf_tk.commands.deploy.read_yaml_file", fake_read_yaml_file)

    monkeypatch.setattr("cognite_toolkit._cdf_tk.utils.load_yaml_inject_variables", fake_load_yaml_inject_variables)
    monkeypatch.setattr(
        "cognite_toolkit._cdf_tk.loaders._base_loaders.load_yaml_inject_variables", fake_load_yaml_inject_variables
    )
    for module in [
        "classic_loaders",
        "datamodel_loaders",
        "industrial_tool_loaders",
        "transformation_loaders",
    ]:
        monkeypatch.setattr(
            f"cognite_toolkit._cdf_tk.loaders._resource_loaders.{module}.load_yaml_inject_variables",
            fake_load_yaml_inject_variables,
        )


class PrintCapture:
    # Find all text within square brackets
    _pattern = re.compile(r"\[([^]]+)\]")

    def __init__(self):
        self.messages = []

    def __call__(
        self,
        *objects: Any,
        sep: str = " ",
        end: str = "\n",
        file: Optional[IO[str]] = None,
        flush: bool = False,
    ):
        for obj in objects:
            if isinstance(obj, str) and (clean := self._pattern.sub("", obj).strip()):
                # Remove square brackets and whitespace. This is to take
                # away the styling from rich print.
                self.messages.append(clean)


T_Type = TypeVar("T_Type", bound=type)


def all_subclasses(base: T_Type) -> list[T_Type]:
    """Returns a list (without duplicates) of all subclasses of a given class, sorted on import-path-name.
    Ignores classes not part of the main library, e.g. subclasses part of tests.
    """
    return sorted(
        filter(
            lambda sub: sub.__module__.startswith("cognite.client"),
            set(base.__subclasses__()).union(s for c in base.__subclasses__() for s in all_subclasses(c)),
        ),
        key=str,
    )


def all_concrete_subclasses(base: T_Type) -> list[T_Type]:
    return [
        sub
        for sub in all_subclasses(base)
        if all(base is not abc.ABC for base in sub.__bases__) and not inspect.isabstract(sub)
    ]


T_Object = TypeVar("T_Object", bound=object)


class FakeCogniteResourceGenerator:
    _error_msg: typing.ClassVar[str] = "Please extend this function to support generating fake data for this type"

    def __init__(
        self,
        seed: int | None = None,
        cognite_client: CogniteClientMock | CogniteClient | None = None,
        max_list_dict_items: int = 3,
    ) -> None:
        self._random = random.Random(seed)
        self._cognite_client = cognite_client or CogniteClientMock()
        self._max_list_dict_items = max_list_dict_items

    def create_instances(self, list_cls: type[T_Object], skip_defaulted_args: bool = False) -> T_Object:
        return list_cls(
            [self.create_instance(list_cls._RESOURCE, skip_defaulted_args) for _ in range(self._max_list_dict_items)]
        )

    def create_instance(self, resource_cls: type[T_Object], skip_defaulted_args: bool = False) -> T_Object:
        is_abstract = any(base is abc.ABC for base in resource_cls.__bases__)
        if is_abstract:
            subclasses = all_concrete_subclasses(resource_cls)
            if not subclasses:
                raise ValueError(f"Cannot create instance of abstract class {resource_cls.__name__}")
            resource_cls = self._random.choice(subclasses)

        signature = inspect.signature(resource_cls.__init__)
        type_hint_by_name = _TypeHints.get_type_hints_by_name(resource_cls)

        keyword_arguments: dict[str, Any] = {}
        positional_arguments: list[Any] = []
        for name, parameter in signature.parameters.items():
            if name == "self":
                continue
            elif name == "args" or name == "kwargs":
                # Skipping generic arguments.
                continue
            elif name.startswith("_"):
                continue
            elif parameter.annotation is inspect.Parameter.empty:
                raise ValueError(f"Parameter {name} of {resource_cls.__name__} is missing annotation")
            elif skip_defaulted_args and parameter.default is not inspect.Parameter.empty:
                continue

            if resource_cls is Geometry and name == "geometries":
                # Special case for Geometry to avoid recursion.
                value = None
            elif name == "scene":
                value = self.create_value(LocationFilterScene, var_name=name)
            elif name == "version":
                # Special case
                value = random.choice(["v1", "v2", "v3"])
            elif resource_cls is LocationFilter and name == "locations":
                # Special case for LocationFilter to avoid recursion.
                value = None
            else:
                value = self.create_value(type_hint_by_name[name], var_name=name)

            if parameter.kind in {parameter.POSITIONAL_ONLY, parameter.VAR_POSITIONAL}:
                positional_arguments.append(value)
            else:
                keyword_arguments[name] = value

        # Special cases
        if resource_cls is DataPointSubscriptionWrite:
            # DataPointSubscriptionWrite requires either timeseries_ids or filter
            if skip_defaulted_args:
                keyword_arguments["time_series_ids"] = ["my_timeseries1", "my_timeseries2"]
            else:
                keyword_arguments.pop("filter", None)

        if resource_cls is Query or resource_cls is WorkflowTriggerDataModelingQuery:
            # The fake generator makes all dicts from 1-3 values, we need to make sure that the query is valid
            # by making sure that the list of equal length, so we make both to length 1.
            with_key, with_value = next(iter(keyword_arguments["with_"].items()))
            select_value = next(iter(keyword_arguments["select"].values()))
            keyword_arguments["with_"] = {with_key: with_value}
            keyword_arguments["select"] = {with_key: select_value}
        elif resource_cls is Relationship and not skip_defaulted_args:
            # Relationship must set the source and target type consistently with the source and target
            keyword_arguments["source_type"] = type(keyword_arguments["source"]).__name__
            keyword_arguments["target_type"] = type(keyword_arguments["target"]).__name__
        elif resource_cls is Datapoints and not skip_defaulted_args:
            # All lists have to be equal in length and only value and timestamp
            keyword_arguments["timestamp"] = keyword_arguments["timestamp"][:1]
            keyword_arguments["value"] = keyword_arguments["value"][:1]
            for key in list(keyword_arguments):
                if isinstance(keyword_arguments[key], list) and key not in {"timestamp", "value"}:
                    keyword_arguments.pop(key)
        elif resource_cls is DatapointsArray:
            keyword_arguments["is_string"] = False
        elif resource_cls is SequenceRows or resource_cls is ToolkitSequenceRows:
            # All row values must match the number of columns
            # Reducing to one column, and one value for each row
            if skip_defaulted_args:
                keyword_arguments["external_id"] = "my_sequence_rows"
            keyword_arguments["columns"] = keyword_arguments["columns"][:1]
            for row in keyword_arguments["rows"]:
                row.values = row.values[:1]
        elif resource_cls is SequenceData:
            if skip_defaulted_args:
                # At least external_id or id must be set
                keyword_arguments["external_id"] = "my_sequence"
                keyword_arguments["rows"] = [
                    SequenceRow(
                        row_number=1,
                        values=[
                            1,
                        ],
                    )
                ]
                keyword_arguments["columns"] = SequenceColumnList(
                    [
                        SequenceColumn("my_column"),
                    ]
                )
            else:
                # All row values must match the number of columns
                keyword_arguments.pop("rows", None)
                keyword_arguments["columns"] = keyword_arguments["columns"][:1]
                keyword_arguments["row_numbers"] = keyword_arguments["row_numbers"][:1]
                keyword_arguments["values"] = keyword_arguments["values"][:1]
                keyword_arguments["values"][0] = keyword_arguments["values"][0][:1]
        elif resource_cls is EndTimeFilter:
            # EndTimeFilter requires either is null or (max and/or min)
            keyword_arguments.pop("is_null", None)
        elif resource_cls is Transformation and not skip_defaulted_args:
            # schedule and jobs must match external id and id
            keyword_arguments["schedule"].external_id = keyword_arguments["external_id"]
            keyword_arguments["schedule"].id = keyword_arguments["id"]
            keyword_arguments["running_job"].transformation_id = keyword_arguments["id"]
            keyword_arguments["last_finished_job"].transformation_id = keyword_arguments["id"]
        elif resource_cls is TransformationScheduleWrite:
            # TransformationScheduleWrite requires either id or external_id
            keyword_arguments.pop("id", None)
        elif resource_cls is TransformationNotificationWrite:
            # TransformationNotificationWrite requires either transformation_id or transformation_external_id
            if skip_defaulted_args:
                keyword_arguments["transformation_external_id"] = "my_transformation"
            else:
                keyword_arguments.pop("transformation_id", None)
        elif resource_cls is NodeResultSetExpression and not skip_defaulted_args:
            # Through has a special format.
            keyword_arguments["through"] = [keyword_arguments["through"][0], "my_view/v1", "a_property"]
        elif resource_cls is RestConfig:
            incremental = keyword_arguments.get("incremental_load")
            if isinstance(incremental, NextUrlLoad):
                # The incremental load cannot be of type `nextUrl`
                load_cls = self._random.choice([BodyLoad, HeaderValueLoad, QueryParamLoad])
                keyword_arguments["incremental_load"] = self.create_instance(load_cls, skip_defaulted_args)
        elif issubclass(resource_cls, ListablePropertyType) and "max_list_size" in keyword_arguments:
            if keyword_arguments["max_list_size"] <= 1:
                keyword_arguments.pop("max_list_size")
            elif keyword_arguments["max_list_size"] > 1:
                keyword_arguments["max_list_size"] = min(2_000, keyword_arguments["max_list_size"])
                keyword_arguments["is_list"] = True

        return resource_cls(*positional_arguments, **keyword_arguments)

    def create_value(self, type_: Any, var_name: str | None = None) -> Any:
        import numpy as np

        if isinstance(type_, typing.ForwardRef):
            type_ = type_._evaluate(globals(), self._type_checking())

        container_type = get_origin(type_)
        is_container = container_type is not None
        if not is_container or container_type is np.ndarray:  # looks weird, but 3.8 and 3.12 type compat. issue
            # Handle numpy types
            from numpy.typing import NDArray

            if type_ == NDArray[np.float64]:
                return np.array([self._random.random() for _ in range(3)], dtype=np.float64)
            elif type_ == NDArray[np.uint32]:
                return np.array([self._random.randint(1, 100) for _ in range(3)], dtype=np.uint32)
            elif type_ == NDArray[np.int64]:
                return np.array([self._random.randint(1, 100) for _ in range(3)], dtype=np.int64)
            elif type_ == NDArray[np.datetime64]:
                return np.array([self._random.randint(1, 1704067200000) for _ in range(3)], dtype="datetime64[ns]")
            elif type_ == NDArray[np.object_]:
                return np.array([self._random_string(10) for _ in range(3)], dtype=np.object_)

        # Handle containers
        args = get_args(type_)
        first_not_none = next((arg for arg in args if arg is not None), None)
        if container_type in UNION_TYPES:
            return self.create_value(first_not_none)
        elif container_type is typing.Literal:
            return self._random.choice(args)
        elif container_type in [
            list,
            typing.Sequence,
            collections.abc.Sequence,
            collections.abc.Collection,
            SequenceNotStr,
        ]:
            return [self.create_value(first_not_none) for _ in range(3)]
        elif container_type in [dict, collections.abc.MutableMapping, collections.abc.Mapping]:
            if first_not_none is None:
                return self.create_value(dict)
            key_type, value_type = args
            return {
                self.create_value(key_type): self.create_value(value_type) for _ in range(self._random.randint(1, 3))
            }
        elif container_type is set:
            return set(self.create_value(first_not_none) for _ in range(self._random.randint(1, 3)))
        elif container_type is tuple:
            if any(arg is ... for arg in args):
                return tuple(self.create_value(first_not_none) for _ in range(self._random.randint(1, 3)))
            raise NotImplementedError(f"Tuple with multiple types is not supported. {self._error_msg}")

        if var_name == "external_id" and type_ is str:
            return self._random_string(50, sample_from=string.ascii_uppercase + string.digits)
        elif var_name == "id" and type_ is int:
            return self._random.choice(range(1, MAX_VALID_INTERNAL_ID + 1))
        if type_ is str or type_ is Any:
            return self._random_string()
        elif type_ is int:
            return self._random.randint(1, 100000)
        elif type_ is float:
            return self._random.random()
        elif type_ is bool:
            return self._random.choice([True, False])
        elif type_ == datetime:
            return datetime.fromtimestamp(self._random.randint(1, 1704067200))
        elif type_ == date:
            return date.fromtimestamp(self._random.randint(1, 1704067200))
        elif type_ is dict:
            return {self._random_string(10): self._random_string(10) for _ in range(self._random.randint(1, 3))}
        elif type_ is CogniteClient:
            return self._cognite_client
        elif inspect.isclass(type_) and any(base is abc.ABC for base in type_.__bases__):
            implementations = all_concrete_subclasses(type_)
            if type_ is Filter:
                # Remove filters not supported by dps subscriptions
                implementations.remove(filters.Overlaps)

                # Remove filters which are only used by data modeling classes
                implementations.remove(filters.HasData)
                implementations.remove(filters.InvalidFilter)
                implementations.remove(filters.Nested)
                implementations.remove(filters.GeoJSONWithin)
                implementations.remove(filters.GeoJSONDisjoint)
                implementations.remove(filters.GeoJSONIntersects)
            elif type_ is Capability:
                implementations.remove(UnknownAcl)
                if LegacyCapability in implementations:
                    implementations.remove(LegacyCapability)
            if type_ is WorkflowTaskOutput:
                # For Workflow Output has to match the input type
                selected = FunctionTaskOutput
            elif type_ is WorkflowTaskParameters:
                selected = FunctionTaskParameters
            else:
                selected = self._random.choice(implementations)
            return self.create_instance(selected)
        elif isinstance(type_, enum.EnumMeta):
            return self._random.choice(list(type_))
        elif isinstance(type_, TypeVar):
            return self.create_value(type_.__bound__)
        elif inspect.isclass(type_) and issubclass(type_, CogniteResourceList):
            return type_([self.create_value(type_._RESOURCE) for _ in range(self._random.randint(1, 3))])
        elif inspect.isclass(type_):
            return self.create_instance(type_)
        elif type(type_) is dataclasses.InitVar:
            return self.create_value(type_.type)

        raise NotImplementedError(f"Unsupported {type_=} or {container_type=}. {self._error_msg}")

    def _random_string(
        self,
        size: int | None = None,
        sample_from: str = string.ascii_uppercase + string.digits + string.ascii_lowercase + string.punctuation,
    ) -> str:
        k = size or self._random.randint(1, 100)
        return "".join(self._random.choices(sample_from, k=k))

    @classmethod
    def _type_checking(cls) -> dict[str, Any]:
        """
        When calling the get_type_hints function, it imports the module with the function TYPE_CHECKING is set to False.

        This function takes all the special types used in data classes and returns them as a dictionary so it
        can be used in the local namespaces.
        """
        import numpy as np
        import numpy.typing as npt
        from cognite.client import CogniteClient

        NumpyDatetime64NSArray = npt.NDArray[np.datetime64]
        NumpyInt64Array = npt.NDArray[np.int64]
        NumpyFloat64Array = npt.NDArray[np.float64]
        NumpyObjArray = npt.NDArray[np.object_]
        return {
            "CogniteClient": CogniteClient,
            "NumpyDatetime64NSArray": NumpyDatetime64NSArray,
            "NumpyInt64Array": NumpyInt64Array,
            "NumpyFloat64Array": NumpyFloat64Array,
            "NumpyObjArray": NumpyObjArray,
        }


class MockQuestion:
    def __init__(self, answer: Any, choices: list[Choice] | None = None) -> None:
        self.answer = answer
        self.choices = choices

    def ask(self) -> Any:
        if isinstance(self.answer, Callable):
            return self.answer(self.choices)
        return self.answer


class MockQuestionary:
    def __init__(self, module_target: str, monkeypatch: MonkeyPatch, answers: list[Any]) -> None:
        self.module_target = module_target
        self.answers = answers
        self.monkeypatch = monkeypatch

    def select(self, _: str, choices: list[Choice] | None = None, **__) -> MockQuestion:
        return MockQuestion(self.answers.pop(0), choices)

    def confirm(self, *_, **__) -> MockQuestion:
        return MockQuestion(self.answers.pop(0))

    def checkbox(self, *_, choices: list[Choice], **__) -> MockQuestion:
        return MockQuestion(self.answers.pop(0), choices)

    def text(self, *_, **__) -> MockQuestion:
        return MockQuestion(self.answers.pop(0))

    def __enter__(self):
        for method in [self.select, self.confirm, self.checkbox, self.text]:
            self.monkeypatch.setattr(f"{self.module_target}.questionary.{method.__name__}", method)
        return self

    def __exit__(self, *args):
        self.monkeypatch.undo()
        return False

    @staticmethod
    def select_all(choices: list[Choice]) -> list[str]:
        if not choices:
            return []
        return [choice.value for choice in choices]


def find_resources(resource: str, resource_dir: str | None = None):
    base = COMPLETE_ORG / MODULES
    for path in base.rglob(f"*{resource}.yaml"):
        if resource_dir and resource_dir not in path.parts:
            continue
        data = read_yaml_file(path)
        if isinstance(data, dict):
            yield pytest.param(data, id=path.relative_to(base).as_posix())
        elif isinstance(data, list):
            for no, item in enumerate(data):
                if isinstance(item, dict):
                    yield pytest.param(item, id=f"{path.relative_to(base).as_posix()} - Item: {no}")
                else:
                    raise ValueError(f"Invalid data format in {path}: {item}")
        else:
            raise ValueError(f"Invalid data format in {path}: {data}")



import pytest
from cognite.client import _version as CogniteSDKVersion
from packaging.requirements import Requirement

from cognite_toolkit._cdf_tk.loaders._resource_loaders.industrial_tool_loaders import StreamlitLoader
from cognite_toolkit._cdf_tk.tk_warnings import StreamlitRequirementsWarning


class TestStreamlitLoader:
    @pytest.fixture
    def recommended_packages(self):
        return StreamlitLoader.recommended_packages()

    def test_requirements(self, recommended_packages):
        assert Requirement("pyodide-http==0.2.1") in recommended_packages
        assert Requirement(f"cognite-sdk=={CogniteSDKVersion.__version__}") in recommended_packages

    def test_requirements_to_str(self, recommended_packages):
        assert set([str(r) for r in recommended_packages]) == set(
            ["pyodide-http==0.2.1", f"cognite-sdk=={CogniteSDKVersion.__version__}"]
        )

    def user_requirements_txt() -> list[tuple[list[str], StreamlitRequirementsWarning]]:
        return [
            (["pyodide-http==0.2.1", "cognite-sdk==7.62.1"], []),
            (["pyodide-http==0.0.0", "cognite-sdk", "numpy"], []),
            (["pyodide-http", "numpy"], ["cognite-sdk"]),
            (["numpy"], ["pyodide-http", "cognite-sdk"]),
            (["numpy", "pandas"], ["pyodide-http", "cognite-sdk"]),
            ([], ["pyodide-http", "cognite-sdk"]),
        ]

    @pytest.mark.parametrize(
        "mock_content, expected_result",
        user_requirements_txt(),
    )
    def test_requirements_file_missing_packages(self, mock_content, expected_result) -> None:
        actual = StreamlitLoader._missing_recommended_requirements(mock_content)
        assert set(actual) == set(expected_result)



from pydantic import Field

from .base import ToolkitResource


class ThreeDModelYAML(ToolkitResource):
    data_set_external_id: str | None = Field(
        default=None, description="The id of the dataset this 3D model belongs to."
    )
    name: str = Field(
        description="The name of the model.",
        min_length=1,
        max_length=255,
    )
    metadata: dict[str, str] | None = Field(
        default=None,
        description="Custom, application-specific metadata.",
        max_length=16,
    )



import csv
import sys
from collections.abc import Collection, Iterator, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import SupportsIndex, overload

from cognite.client.data_classes.data_modeling import NodeId

from cognite_toolkit._cdf_tk.client.data_classes.pending_instances_ids import PendingInstanceId
from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitFileNotFoundError,
    ToolkitValueError,
)

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


@dataclass
class MigrationMapping:
    resource_type: str
    instance_id: NodeId


@dataclass
class IdMigrationMapping(MigrationMapping):
    id: int
    data_set_id: int | None = None


@dataclass
class ExternalIdMigrationMapping(MigrationMapping):
    external_id: str
    data_set_id: int | None = None


class MigrationMappingList(list, Sequence[MigrationMapping]):
    # Implemented to get correct type hints
    def __init__(self, collection: Collection[MigrationMapping] | None = None) -> None:
        super().__init__(collection or [])

    def __iter__(self) -> Iterator[MigrationMapping]:
        return super().__iter__()

    @overload
    def __getitem__(self, index: SupportsIndex) -> MigrationMapping: ...

    @overload
    def __getitem__(self, index: slice) -> "MigrationMappingList": ...

    def __getitem__(self, index: SupportsIndex | slice, /) -> "MigrationMapping | MigrationMappingList":
        if isinstance(index, slice):
            return MigrationMappingList(super().__getitem__(index))
        return super().__getitem__(index)

    def get_ids(self) -> list[int]:
        """Return a list of IDs from the migration mappings."""
        return [mapping.id for mapping in self if isinstance(mapping, IdMigrationMapping)]

    def get_external_ids(self) -> list[str]:
        """Return a list of external IDs from the migration mappings."""
        return [mapping.external_id for mapping in self if isinstance(mapping, ExternalIdMigrationMapping)]

    def as_node_ids(self) -> list[NodeId]:
        """Return a list of NodeIds from the migration mappings."""
        return [mapping.instance_id for mapping in self]

    def spaces(self) -> set[str]:
        """Return a set of spaces from the migration mappings."""
        return {mapping.instance_id.space for mapping in self}

    def as_pending_ids(self) -> list[PendingInstanceId]:
        return [
            PendingInstanceId(
                pending_instance_id=mapping.instance_id,
                id=mapping.id if isinstance(mapping, IdMigrationMapping) else None,
                external_id=mapping.external_id if isinstance(mapping, ExternalIdMigrationMapping) else None,
            )
            for mapping in self
        ]

    def get_data_set_ids(self) -> set[int]:
        """Return a list of data set IDs from the migration mappings."""
        return {
            mapping.data_set_id
            for mapping in self
            if isinstance(mapping, IdMigrationMapping | ExternalIdMigrationMapping) and mapping.data_set_id is not None
        }

    @classmethod
    def read_mapping_file(cls, mapping_file: Path) -> Self:
        if not mapping_file.exists():
            raise ToolkitFileNotFoundError(f"Mapping file {mapping_file} does not exist.")
        if mapping_file.suffix != ".csv":
            raise ToolkitValueError(f"Mapping file {mapping_file} must be a CSV file.")
        with mapping_file.open(mode="r") as f:
            csv_file = csv.reader(f)
            header = next(csv_file, None)
            header = cls._validate_csv_header(header)
            if header[0] == "id":
                return cls._read_id_mapping(csv_file)
            else:  # header[0] == "externalId":
                return cls._read_external_id_mapping(csv_file)

    @classmethod
    def _validate_csv_header(cls, header: list[str] | None) -> list[str]:
        if header is None:
            raise ToolkitValueError("Mapping file is empty")
        errors: list[str] = []
        if len(header) < 3:
            errors.append("Mapping file must have at least 3 columns: id/externalId, space, externalId.")
        if len(header) >= 5:
            errors.append("Mapping file must have at most 4 columns: id/externalId, dataSetId, space, externalId.")
        if header[0] not in {"id", "externalId"}:
            errors.append("First column must be 'id' or 'externalId'.")
        if len(header) == 4 and header[1] != "dataSetId":
            errors.append("If there are 4 columns, the second column must be 'dataSetId'.")
        if header[-2:] != ["space", "externalId"]:
            errors.append("Last two columns must be 'space' and 'externalId'.")
        if errors:
            error_str = "\n - ".join(errors)
            raise ToolkitValueError(f"Invalid mapping file header:\n - {error_str}")
        return header

    @classmethod
    def _read_id_mapping(cls, csv_file: Iterator[list[str]]) -> Self:
        """Read a CSV file with ID mappings."""
        mappings = cls()
        for no, row in enumerate(csv_file, 1):
            try:
                id_ = int(row[0])
                data_set_id = int(row[1]) if len(row) == 4 and row[1] else None
            except ValueError as e:
                raise ToolkitValueError(
                    f"Invalid ID or dataSetId in row {no}: {row}. ID and dataSetId must be integers."
                ) from e
            instance_id = NodeId(*row[-2:])
            mappings.append(
                IdMigrationMapping(resource_type="timeseries", id=id_, instance_id=instance_id, data_set_id=data_set_id)
            )
        return mappings

    @classmethod
    def _read_external_id_mapping(cls, csv_file: Iterator[list[str]]) -> Self:
        """Read a CSV file with external ID mappings."""
        mappings = cls()
        for row in csv_file:
            external_id = row[0]
            data_set_id = int(row[1]) if len(row) == 4 and row[1] else None
            instance_id = NodeId(*row[-2:])
            mappings.append(
                ExternalIdMigrationMapping(
                    resource_type="timeseries",
                    external_id=external_id,
                    instance_id=instance_id,
                    data_set_id=data_set_id,
                )
            )
        return mappings



import pytest
from cognite.client.data_classes.data_modeling import Space
from cognite.client.data_classes.data_modeling.cdm.v1 import CogniteTimeSeriesApply

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.instances import InstancesApplyResultList
from tests.test_integration.constants import RUN_UNIQUE_ID


@pytest.fixture
def toolkit_client_create_limit_1(toolkit_client: ToolkitClient) -> ToolkitClient:
    before = toolkit_client.data_modeling.instances._CREATE_LIMIT
    toolkit_client.data_modeling.instances._CREATE_LIMIT = 1
    yield toolkit_client
    toolkit_client.data_modeling.instances._CREATE_LIMIT = before


class TestExtendedInstancesAPI:
    def test_apply_fast_instances(self, toolkit_client_create_limit_1: ToolkitClient, toolkit_space: Space) -> None:
        client = toolkit_client_create_limit_1
        ts = [
            CogniteTimeSeriesApply(
                space=toolkit_space.space,
                external_id=f"toolkit_test_apply_fast_{i}_{RUN_UNIQUE_ID}",
                is_step=False,
                time_series_type="numeric",
                name=f"Toolkit Test Time Series {i}",
            )
            for i in range(4)
        ]
        created: InstancesApplyResultList | None = None
        try:
            created = client.data_modeling.instances.apply_fast(ts)
            assert len(created) == 4
        finally:
            if created is not None:
                client.data_modeling.instances.delete(created.as_ids())



"""360 image extractor base class."""

import io
import logging
import time
from dataclasses import asdict, dataclass
from enum import Enum
from typing import Any, Union

import numpy as np
import py360convert
from cognite.client.data_classes import Event, FileMetadata, Label
from PIL import Image

logger = logging.getLogger(__name__)

EVENT_TYPE = "scan"
EVENT_SUB_TYPE = "terrestial"
CUBEMAP_RESOLUTION = 2048


@dataclass
class VectorXYZ:
    """Vector of floats:  (x,y,z)."""

    x: float = 0
    y: float = 0
    z: float = 0

    def to_string(self) -> str:
        """Convert vector so string."""
        return f"{self.x:.4f}, {self.y:.4f}, {self.z:.4f}"


def translation_to_mm_str_with_offset(translation: VectorXYZ, translation_offset_mm: VectorXYZ, translation_unit: str):
    """Transfrom translation to string in to mm with offset."""
    unit_clean = translation_unit.lower().replace(" ", "")
    if unit_clean == "m":
        scale = 1000
    elif unit_clean == "cm":
        scale = 10
    elif unit_clean == "mm":
        scale = 1
    else:
        raise ValueError(
            f"Translaiton unit not recognized: {translation_unit}. " 'Translation unit should be "m", "cm" or "mm".'
        )

    return VectorXYZ(
        x=(translation.x * scale - translation_offset_mm.x),
        y=(translation.y * scale - translation_offset_mm.y),
        z=(translation.z * scale - translation_offset_mm.z),
    ).to_string()


def convert_rotation_angle(rotation_angle: str, rotation_angle_unit: str) -> str:
    """Check rotation angle unit and convert to radians."""
    if rotation_angle_unit == "deg":
        return rotation_angle
    elif rotation_angle_unit == "rad":
        return str(np.rad2deg(float(rotation_angle)))
    else:
        raise ValueError(
            f"Rotation angle unit not recognized: {rotation_angle_unit}. "
            'Rotation angle unit should be "rad" or "deg".'
        )


@dataclass
class ThreesixtyImageMetadata:
    """ThreesixtyImage metadata. The metadata of the 360 events."""

    station_id: str
    station_name: str
    rotation_angle: str  # Format: "14.837"
    rotation_axis: str  # Format: "0.0990, 0.0113, 0.9950"
    translation: str  # Format: "343109.0000, 83408.0000, 30860.0000"
    timestamp: int  # Format: milliseconds since epoch
    site_id: str = ""
    site_name: str = ""


@dataclass
class ThreesixtyImage:
    """All information about the station."""

    station_number: str
    tran_unit: str
    rot_angle_unit: str
    images: dict[str, Union[str, np.ndarray]]
    threesixty_image_metadata: ThreesixtyImageMetadata


@dataclass
class ImageWithFileMetadata:
    """Image with file metadata."""

    file_metadata: FileMetadata
    content: bytes


class CogniteThreeSixtyImageExtractor:
    """Base class for 360 image extractors.

    360 images are represented by ThreesixtyImage objects. The extractor use the station objects
    to create 360 files and events in CDF.
    """

    def __init__(self, data_set_id: int, mime_type: str = "image/jpeg"):
        """Initialize ThreeSixtyImageExtractor."""
        self.data_set_id: int = data_set_id
        self.mime_type = mime_type
        self.labels: list[Label] = []

    class Faces(Enum):
        left = 0
        front = 1
        right = 2
        back = 3
        top = 4
        bottom = 5

    def create_threesixty_image(
        self,
        content: np.ndarray,
        site_id: str,
        site_name: str,
        station_number: Any,
        rotation_angle: str,
        rotation_axis: VectorXYZ,
        rotation_angle_unit: str,
        translation: VectorXYZ,
        translation_unit: str,
        translation_offset_mm: VectorXYZ,
        timestamp: int = 0,
    ) -> tuple[Event, list[ImageWithFileMetadata]]:
        """Append station measurement to station list for truview."""
        if timestamp == 0 or timestamp is None:
            timestamp = int(time.time() * 1000)
        if not isinstance(timestamp, int):
            timestamp = int(timestamp)
        station = ThreesixtyImage(
            station_number=station_number,
            rot_angle_unit="deg",
            tran_unit="mm",
            images=self._get_cubemap_images(content),
            threesixty_image_metadata=ThreesixtyImageMetadata(
                site_id=site_id,
                site_name=site_name,
                station_name=site_name + " " + station_number,
                rotation_angle=convert_rotation_angle(rotation_angle, rotation_angle_unit),
                rotation_axis=rotation_axis.to_string(),
                translation=translation_to_mm_str_with_offset(translation, translation_offset_mm, translation_unit),
                station_id=site_id + "-" + station_number,
                timestamp=timestamp,
            ),
        )

        event = self._create_cdf_events(station)
        files = self._create_cdf_files(station)
        return event, files

    def _create_cdf_events(self, three_sixty_image: ThreesixtyImage) -> Event:
        """Create 360 image eventfrom

        One event per 360 image is created.
        """
        logger.info("Create events.")
        event = Event(
            three_sixty_image.threesixty_image_metadata.station_id
            + str(three_sixty_image.threesixty_image_metadata.timestamp)
        )
        event.metadata = asdict(three_sixty_image.threesixty_image_metadata)
        event.data_set_id = self.data_set_id
        event.description = "Scan position " + three_sixty_image.threesixty_image_metadata.station_name
        event.type = EVENT_TYPE
        event.subtype = EVENT_SUB_TYPE
        event.start_time = three_sixty_image.threesixty_image_metadata.timestamp
        return event

    def _image_to_byte_array(self, image: Image):
        img_bytes = io.BytesIO()
        image.save(img_bytes, format="PNG")
        img_bytes = img_bytes.getvalue()  # type: ignore
        return img_bytes

    def _create_cdf_files(
        self, three_sixty_image: ThreesixtyImage, resolution: int = 2048
    ) -> list[ImageWithFileMetadata]:
        """Create 360 image files from three sixty image.

        Six files per 360 image is created, one file per face.
        """
        logger.info("Create 360 files.")
        files = []
        if three_sixty_image.threesixty_image_metadata.station_id:
            for i, face in enumerate(self.Faces):
                file_metadata = FileMetadata(
                    three_sixty_image.threesixty_image_metadata.station_id
                    + str(three_sixty_image.threesixty_image_metadata.timestamp)
                    + "-"
                    + str(resolution)
                    + "-"
                    + face.name
                )
                file_metadata.labels = self.labels
                metadata = {}
                metadata["site_id"] = three_sixty_image.threesixty_image_metadata.site_id
                metadata["site_name"] = three_sixty_image.threesixty_image_metadata.site_name
                metadata["station_id"] = three_sixty_image.threesixty_image_metadata.station_id
                metadata["station_name"] = three_sixty_image.threesixty_image_metadata.station_name
                metadata["timestamp"] = str(three_sixty_image.threesixty_image_metadata.timestamp)
                metadata["image_type"] = "cubemap"
                metadata["image_resolution"] = str(resolution)
                metadata["face"] = face.name
                metadata["processed"] = "false"
                file_metadata.metadata = metadata
                file_metadata.name = (
                    three_sixty_image.threesixty_image_metadata.station_name
                    + "-"
                    + str(three_sixty_image.threesixty_image_metadata.timestamp)  # For avoiding duplicate filenames
                    + "-"
                    + face.name
                    + ".jpg"
                )
                file_metadata.mime_type = self.mime_type
                file_metadata.data_set_id = self.data_set_id
                cubemap_img = three_sixty_image.images[face.name]
                content = self._image_to_byte_array(Image.fromarray(cubemap_img))
                files.append(ImageWithFileMetadata(content=content, file_metadata=file_metadata))
        else:
            logger.error("No station ID found.")
        return files

    def _cube_dice2h(self, cube_dice):
        w = cube_dice.shape[0] // 3
        assert cube_dice.shape[0] == w * 3 and cube_dice.shape[1] == w * 4
        cube_h = np.zeros((w, w * 6, cube_dice.shape[2]), dtype=cube_dice.dtype)
        # Order: F R B L U D
        sxy = [(1, 1), (2, 1), (3, 1), (0, 1), (1, 0), (1, 2)]
        for i, (sx, sy) in enumerate(sxy):
            face = cube_dice[(sy * w) : (sy + 1) * w, (sx * w) : (sx + 1) * w]
            cube_h[:, (i * w) : (i + 1) * w] = face
        return cube_h

    def _get_cubemap_images(self, content: Union[str, np.ndarray]) -> Union[dict[str, str], dict[str, np.ndarray]]:
        """Create cubemap dict from equirectangular image.

        Args:
        content: equirectangular
        Returns:
        cubemaps (dict[str,np.ndarray]: {<face>: image})
        """
        if not isinstance(content, np.ndarray):
            logger.error(
                f"Unsupported input type: Equirectangular extractor"
                f"only supports np.ndarray images. Got type {type(content)}"
            )
            raise TypeError(
                f"Unsupported input type: Equirectangular extractor "
                f"only supports np.ndarray images. Got type {type(content)}"
            )

        cubemaps: dict[str, np.ndarray] = {}
        try:
            logger.info("Creating cubemap images from equirectangular image.")
            im = py360convert.e2c(content, face_w=CUBEMAP_RESOLUTION)
            cube_h = self._cube_dice2h(im)
            cube_dict = py360convert.cube_h2dict(cube_h)

            # Translate keys from py360convert to TruView naming convention.
            dice_map = {
                "F": "front",
                "R": "right",
                "B": "back",
                "L": "left",
                "U": "top",
                "D": "bottom",
            }
            for key, im in cube_dict.items():
                cubemaps[dice_map[key]] = cube_dict[key]
        except Exception as e:
            raise Exception(f"Failed to create cubemap image from equirectangular image: {e}.")
        return cubemaps
